#  Copyright © 2020 Ingram Micro Inc. All rights reserved.


class BaseTransport:
    """
    CQRS pattern can be implemented over any transport (AMQP, HTTP, etc.)
    All transports need to inherit from this base class.
    Transport must be set in Django settings:
        CQRS = {
            'transport': 'dj_cqrs.transport.rabbit_mq.RabbitMQTransport',
        }
    """
    consumers = {}

    @staticmethod
    def produce(payload):
        """ Send data from master model to replicas.

        :param dj_cqrs.dataclasses.TransportPayload payload: Transport payload from master model.
        """
        raise NotImplementedError

    @staticmethod
    def consume(*args, **kwargs):
        raise NotImplementedError



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import os

from dj_cqrs.controller import consumer
from dj_cqrs.transport.base import BaseTransport
from dj_cqrs.transport.rabbit_mq import RabbitMQTransport


class TransportStub(BaseTransport):
    @staticmethod
    def produce(payload):
        TransportStub.consume(payload)

    @staticmethod
    def consume(payload):
        consumer.consume(payload)


class RabbitMQTransportWithEvents(RabbitMQTransport):
    @staticmethod
    def _log_consumed(payload):
        from tests.dj_replica.models import Event
        Event.objects.create(
            pid=os.getpid(),
            cqrs_id=payload.cqrs_id,
            cqrs_revision=int(payload.instance_data['cqrs_revision']),
        )



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

from django.db import models, transaction
from django.dispatch import Signal
from django.utils.timezone import now

from dj_cqrs.controller import producer
from dj_cqrs.constants import SignalType
from dj_cqrs.dataclasses import TransportPayload

post_bulk_create = Signal(providing_args=['instances', 'using'])
post_update = Signal(providing_args=['instances', 'using'])


class MasterSignals:
    """ Signals registry and handlers for CQRS master models. """
    @classmethod
    def register_model(cls, model_cls):
        """
        :param dj_cqrs.mixins.MasterMixin model_cls: Class inherited from CQRS MasterMixin.
        """
        models.signals.post_save.connect(cls.post_save, sender=model_cls)
        models.signals.post_delete.connect(cls.post_delete, sender=model_cls)

        post_bulk_create.connect(cls.post_bulk_create, sender=model_cls)
        post_update.connect(cls.post_bulk_update, sender=model_cls)

    @classmethod
    def post_save(cls, sender, **kwargs):
        """
        :param dj_cqrs.mixins.MasterMixin sender: Class or instance inherited from CQRS MasterMixin.
        """
        if not sender.CQRS_PRODUCE:
            return

        instance = kwargs['instance']
        if not instance.is_sync_instance():
            return

        using = kwargs['using']

        sync = kwargs.get('sync', False)
        queue = kwargs.get('queue', None)

        if not transaction.get_connection(using).in_atomic_block:
            instance_data = instance.to_cqrs_dict(using)

            signal_type = SignalType.SYNC if sync else SignalType.SAVE
            payload = TransportPayload(
                signal_type, sender.CQRS_ID, instance_data, instance.pk, queue,
            )
            producer.produce(payload)

        else:
            transaction.on_commit(
                lambda: MasterSignals.post_save(
                    sender, instance=instance, using=using, sync=sync, queue=queue,
                ),
            )

    @classmethod
    def post_delete(cls, sender, **kwargs):
        """
        :param dj_cqrs.mixins.MasterMixin sender: Class or instance inherited from CQRS MasterMixin.
        """
        if not sender.CQRS_PRODUCE:
            return

        instance = kwargs['instance']
        if not instance.is_sync_instance():
            return

        instance_data = {
            'id': instance.pk,
            'cqrs_revision': instance.cqrs_revision + 1,
            'cqrs_updated': str(now()),
        }
        signal_type = SignalType.DELETE

        payload = TransportPayload(signal_type, sender.CQRS_ID, instance_data, instance.pk)
        # Delete is always in transaction!
        transaction.on_commit(lambda: producer.produce(payload))

    @classmethod
    def post_bulk_create(cls, sender, **kwargs):
        """
        :param dj_cqrs.mixins.MasterMixin sender: Class or instance inherited from CQRS MasterMixin.
        """
        for instance in kwargs['instances']:
            cls.post_save(sender, instance=instance, using=kwargs['using'])

    @classmethod
    def post_bulk_update(cls, sender, **kwargs):
        """
        :param dj_cqrs.mixins.MasterMixin sender: Class or instance inherited from CQRS MasterMixin.
        """
        for instance in kwargs['instances']:
            cls.post_save(sender, instance=instance, using=kwargs['using'])



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import six
from django.db import DatabaseError


def assert_is_sub_dict(dict1, dict2):
    assert six.viewitems(dict1) <= six.viewitems(dict2)


def assert_publisher_once_called_with_args(publisher_mock, *args):
    assert publisher_mock.call_count == 1
    call_t_payload = publisher_mock.call_args[0][0]

    assert call_t_payload.signal_type == args[0]
    assert call_t_payload.cqrs_id == args[1]
    assert call_t_payload.pk == args[3]
    assert_is_sub_dict(args[2], call_t_payload.instance_data)

    required_fields = {'cqrs_revision', 'cqrs_updated'}
    assert not (required_fields - set(call_t_payload.instance_data.keys()))


def db_error(*args, **kwargs):
    raise DatabaseError()



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import pytest
from django.core.management import CommandError, call_command
from django.utils.timezone import now
from tests.utils import db_error

from tests.dj_replica.models import AuthorRef


COMMAND_NAME = 'cqrs_bulk_load'
DUMPS_PATH = 'tests/test_commands/dumps/'


def test_no_input():
    with pytest.raises(CommandError):
        call_command(COMMAND_NAME)


def test_no_file():
    with pytest.raises(CommandError) as e:
        call_command(COMMAND_NAME, '-i=__init__1.py')

    assert "File __init__1.py doesn't exist!" in str(e)


def test_empty_file():
    with pytest.raises(CommandError) as e:
        call_command(COMMAND_NAME, '-i={}empty_file.dump'.format(DUMPS_PATH))

    assert "empty_file.dump is empty!" in str(e)


def test_no_cqrs_id():
    with pytest.raises(CommandError) as e:
        call_command(COMMAND_NAME, '-i={}bad_cqrs_id.dump'.format(DUMPS_PATH))

    assert "Wrong CQRS ID: publisher!" in str(e)


@pytest.mark.django_db
def test_unparseable_line(capsys):
    call_command(COMMAND_NAME, '-i={}unparseable.dump'.format(DUMPS_PATH))
    assert AuthorRef.objects.count() == 0

    captured = capsys.readouterr()
    assert "Dump file can't be parsed: line 2!" in captured.err
    assert '0 instance(s) loaded.' in captured.err


@pytest.mark.django_db
def test_bad_master_data(capsys):
    call_command(COMMAND_NAME, '-i={}bad_master_data.dump'.format(DUMPS_PATH))
    assert AuthorRef.objects.count() == 1

    captured = capsys.readouterr()
    assert "Instance can't be saved: line 3!" in captured.err
    assert '1 instance(s) loaded.' in captured.err


@pytest.mark.django_db
def test_no_rows(capsys):
    AuthorRef.objects.create(id=1, name='1', cqrs_revision=0, cqrs_updated=now())

    call_command(COMMAND_NAME, '--input={}no_rows.dump'.format(DUMPS_PATH))
    assert AuthorRef.objects.count() == 1

    captured = capsys.readouterr()
    assert '0 instance(s) loaded.' in captured.err


@pytest.mark.django_db(transaction=True)
def test_loaded_correctly(capsys):
    AuthorRef.objects.create(id=1, name='1', cqrs_revision=0, cqrs_updated=now())

    call_command(COMMAND_NAME, '--input={}author.dump'.format(DUMPS_PATH))
    assert AuthorRef.objects.count() == 2

    captured = capsys.readouterr()
    assert '2 instance(s) loaded.' in captured.err


@pytest.mark.django_db
def test_delete_before_upload_ok(capsys):
    AuthorRef.objects.create(id=1, name='1', cqrs_revision=0, cqrs_updated=now())

    call_command(COMMAND_NAME, '--input={}no_rows.dump'.format(DUMPS_PATH), '--clear=true')
    assert AuthorRef.objects.count() == 0

    captured = capsys.readouterr()
    assert '0 instance(s) loaded.' in captured.err


@pytest.mark.django_db
def test_delete_operation_fails(mocker, ):
    mocker.patch('django.db.models.manager.BaseManager.all', side_effect=db_error)
    with pytest.raises(CommandError) as e:
        call_command(COMMAND_NAME, '--input={}no_rows.dump'.format(DUMPS_PATH), '--clear=true')

    assert "Delete operation fails!" in str(e)


@pytest.mark.django_db
def test_unexpected_error(mocker, capsys):
    mocker.patch('tests.dj_replica.models.AuthorRef.cqrs_save', side_effect=db_error)
    call_command(COMMAND_NAME, '--input={}author.dump'.format(DUMPS_PATH))

    captured = capsys.readouterr()
    assert 'Unexpected error: line 2!' in captured.err
    assert 'Unexpected error: line 3!' in captured.err
    assert '0 instance(s) loaded.' in captured.err


@pytest.mark.django_db(transaction=True)
def test_loaded_correctly_batch(capsys):
    AuthorRef.objects.create(id=1, name='1', cqrs_revision=0, cqrs_updated=now())

    call_command(COMMAND_NAME, '--input={}author.dump'.format(DUMPS_PATH), '--batch=1')
    assert AuthorRef.objects.count() == 2

    captured = capsys.readouterr()
    assert '2 instance(s) loaded.' in captured.err



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import pytest
from django.db import transaction

from tests.dj_master import models as master_models
from tests.dj_replica import models as replica_models


@pytest.mark.django_db(transaction=True)
def test_create():
    master_model = master_models.BasicFieldsModel.objects.create(
        int_field=1,
        char_field='text',
    )
    assert replica_models.BasicFieldsModelRef.objects.count() == 1

    replica_model = replica_models.BasicFieldsModelRef.objects.first()
    for field_name in ('cqrs_revision', 'cqrs_updated', 'int_field', 'char_field'):
        assert getattr(master_model, field_name) == getattr(replica_model, field_name)


@pytest.mark.django_db(transaction=True)
def test_update():
    master_model = master_models.BasicFieldsModel.objects.create(
        int_field=1,
        char_field='text',
    )
    assert replica_models.BasicFieldsModelRef.objects.count() == 1

    master_model.char_field = 'new_text'
    master_model.save()
    master_model.refresh_from_db()

    assert replica_models.BasicFieldsModelRef.objects.count() == 1

    replica_model = replica_models.BasicFieldsModelRef.objects.first()
    for field_name in ('cqrs_revision', 'cqrs_updated', 'int_field', 'char_field'):
        assert getattr(master_model, field_name) == getattr(replica_model, field_name)


@pytest.mark.django_db(transaction=True)
def test_delete():
    master_model = master_models.BasicFieldsModel.objects.create(
        int_field=1,
        char_field='text',
    )
    assert replica_models.BasicFieldsModelRef.objects.count() == 1

    master_model.delete()
    assert replica_models.BasicFieldsModelRef.objects.count() == 0


@pytest.mark.django_db(transaction=True)
def test_create_serialized():
    with transaction.atomic():
        publisher = master_models.Publisher.objects.create(id=1, name='publisher')
        author = master_models.Author.objects.create(id=1, name='author', publisher=publisher)
        for index in range(1, 3):
            master_models.Book.objects.create(id=index, title=str(index), author=author)

    replica_publisher = replica_models.Publisher.objects.first()
    replica_author = replica_models.AuthorRef.objects.first()
    books = list(replica_models.Book.objects.all())

    assert replica_publisher.id == 1
    assert replica_publisher.name == 'publisher'

    assert replica_author.id == 1
    assert replica_author.name == 'author'
    assert replica_author.publisher == replica_publisher
    assert replica_author.cqrs_revision == 0

    assert len(books) == 2
    assert {1} == {book.author_id for book in books}
    assert {1, 2} == {book.id for book in books}


@pytest.mark.django_db(transaction=True)
def test_update_serialized():
    author = master_models.Author.objects.create(id=1, name='author')
    author.refresh_from_db()

    assert replica_models.AuthorRef.objects.count() == 1

    with transaction.atomic():
        publisher = master_models.Publisher.objects.create(id=1, name='publisher')

        author.publisher = publisher
        author.save()

    replica_publisher = replica_models.Publisher.objects.first()
    replica_author = replica_models.AuthorRef.objects.first()

    assert replica_publisher.id == 1
    assert replica_publisher.name == 'publisher'

    assert replica_author.id == 1
    assert replica_author.name == 'author'
    assert replica_author.publisher == replica_publisher
    assert replica_author.cqrs_revision == 1


@pytest.mark.django_db(transaction=True)
def test_delete_serialized():
    with transaction.atomic():
        author = master_models.Author.objects.create(id=1, name='author')
        for index in range(1, 3):
            master_models.Book.objects.create(id=index, title=str(index), author=author)

    assert replica_models.AuthorRef.objects.count() == 1
    assert replica_models.Book.objects.count() == 2

    author.refresh_from_db()
    author.delete()

    assert replica_models.AuthorRef.objects.count() == 0
    assert replica_models.Book.objects.count() == 0


@pytest.mark.django_db(transaction=True)
def test_sync_is_deleted():
    author = master_models.Author.objects.create(id=1, name='author')
    author.refresh_from_db()

    assert replica_models.AuthorRef.objects.count() == 1

    replica_models.AuthorRef.objects.all().delete()
    author.cqrs_sync()

    assert replica_models.AuthorRef.objects.count() == 1


@pytest.mark.django_db(transaction=True)
def test_sync_exists(mocker):
    author = master_models.Author.objects.create(id=1, name='author')
    author.refresh_from_db()

    assert replica_models.AuthorRef.objects.count() == 1

    # We simulate transport error
    mocker.patch('dj_cqrs.controller.producer.produce')
    author.name = 'new'
    author.save()
    mocker.stopall()

    author.cqrs_sync()

    assert replica_models.AuthorRef.objects.count() == 1
    replica_author = replica_models.AuthorRef.objects.first()
    assert replica_author.cqrs_revision == 1
    assert replica_author.name == 'new'


@pytest.mark.django_db(transaction=True)
def test_sync_downgrade(mocker, caplog):
    author = master_models.Author.objects.create(id=1, name='author')
    author.name = 'new'
    author.save()

    assert replica_models.AuthorRef.objects.count() == 1
    replica_author = replica_models.AuthorRef.objects.first()
    assert replica_author.cqrs_revision == 1
    assert replica_author.name == 'new'

    mocker.patch('dj_cqrs.controller.producer.produce')
    author.delete()
    author = master_models.Author.objects.create(id=1, name='other')
    mocker.stopall()

    author.cqrs_sync()

    assert replica_models.AuthorRef.objects.count() == 1
    replica_author = replica_models.AuthorRef.objects.first()
    assert replica_author.cqrs_revision == 0
    assert replica_author.name == 'other'

    assert 'CQRS revision downgrade on sync: pk = 1, cqrs_revision = new 0 / existing 1 (author).' \
        in caplog.text



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import psycopg2
import pytest

from integration_tests.tests.utils import REPLICA_TABLES


@pytest.fixture
def replica_cursor():
    connection = psycopg2.connect(host='postgres', database='replica', user='user', password='pswd')

    # That setting is extremely important for table truncations
    connection.set_isolation_level(0)

    cursor = connection.cursor()
    for table in REPLICA_TABLES:
        cursor.execute('TRUNCATE TABLE {};'.format(table))

    yield cursor

    cursor.close()
    connection.close()



#!/usr/bin/env python

#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import os
import sys

if __name__ == "__main__":
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "replica_settings")

    from django.core.management import execute_from_command_line
    execute_from_command_line(sys.argv)



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import sys

from django.core.management.base import BaseCommand, CommandError

from dj_cqrs.constants import NO_QUEUE
from dj_cqrs.management.commands.cqrs_sync import Command as SyncCommand
from dj_cqrs.registries import MasterRegistry


class Command(BaseCommand):
    help = 'Diff synchronizer from CQRS replica stream.'

    def handle(self, *args, **options):
        with sys.stdin as f:
            first_line = f.readline().strip()
            model = self._get_model(first_line)
            queue = self._get_queue(first_line)

            for pks_line in f:
                sync_kwargs = {
                    'cqrs_id': model.CQRS_ID,
                    'filter': '{{"id__in": {}}}'.format(pks_line.strip()),
                }
                if queue:
                    sync_kwargs['queue'] = queue

                SyncCommand().handle(**sync_kwargs)

    @staticmethod
    def _get_model(first_line):
        cqrs_id = first_line.split(',')[0]
        model = MasterRegistry.get_model_by_cqrs_id(cqrs_id)

        if not model:
            raise CommandError('Wrong CQRS ID: {}!'.format(cqrs_id))

        return model

    @staticmethod
    def _get_queue(first_line):
        queue = first_line.split(',')[-1]
        if queue != NO_QUEUE:
            return queue



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

from multiprocessing import Process

from django.core.management.base import BaseCommand

from dj_cqrs.transport import current_transport


class Command(BaseCommand):
    help = 'Starts CQRS worker, which consumes messages from message queue.'

    def add_arguments(self, parser):
        parser.add_argument('--workers', '-w', help='Number of workers', type=int, default=0)

    def handle(self, *args, **options):
        if options['workers'] == 0:
            current_transport.consume()
        else:
            pool = []

            for i in range(options['workers']):
                p = Process(target=current_transport.consume)
                pool.append(p)
                p.start()

            for p in pool:
                p.join()



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import pytest
from django.core.management import call_command
from django.db import transaction

from tests.dj_master import models as master_models
from tests.dj_replica import models as replica_models
from tests.test_commands.utils import remove_file


@pytest.mark.django_db
def tests_dumps_several_rows(mocker):
    mocker.patch('dj_cqrs.controller.producer.produce')
    remove_file('bulk_flow.dump')

    master_models.Author.objects.create(id=2, name='2')

    with transaction.atomic():
        publisher = master_models.Publisher.objects.create(id=1, name='publisher')
        master_models.Author.objects.create(id=1, name='1', publisher=publisher)

    assert replica_models.AuthorRef.objects.count() == 0
    assert replica_models.Publisher.objects.count() == 0

    call_command('cqrs_bulk_dump', '--cqrs-id=author', '-o=bulk_flow.dump')
    call_command('cqrs_bulk_load', '-i=bulk_flow.dump')

    assert replica_models.AuthorRef.objects.count() == 2
    assert replica_models.Publisher.objects.count() == 1



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

from importlib import import_module

import pytest
from six.moves import reload_module

from dj_cqrs.transport.base import BaseTransport


def test_no_transport_setting(settings):
    settings.CQRS = {}

    with pytest.raises(AttributeError) as e:
        reload_module(import_module('dj_cqrs.transport'))

    assert str(e.value) == 'CQRS transport is not set.'


def test_bad_transport_setting(settings):
    settings.CQRS = {'transport': '1221'}

    with pytest.raises(ImportError) as e:
        reload_module(import_module('dj_cqrs.transport'))

    assert str(e.value) == 'Bad CQRS transport class.'


class NoneBaseTransportCls(object):
    pass


def test_not_inherited_from_base_transport(settings):
    settings.CQRS = {'transport': 'tests.test_transport.test_base.NoneBaseTransportCls'}

    with pytest.raises(ImportError) as e:
        reload_module(import_module('dj_cqrs.transport'))

    assert str(e.value) == 'Bad CQRS transport class.'


def test_base_transport_consume():
    with pytest.raises(NotImplementedError):
        BaseTransport.consume(None)


def test_base_transport_produce():
    with pytest.raises(NotImplementedError):
        BaseTransport.produce(None)



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.


class TransportPayload:
    def __init__(self, signal_type, cqrs_id, instance_data, instance_pk, queue=None):
        self.__signal_type = signal_type
        self.__cqrs_id = cqrs_id
        self.__instance_data = instance_data
        self.__instance_pk = instance_pk
        self.__queue = queue

    @property
    def signal_type(self):
        return self.__signal_type

    @property
    def cqrs_id(self):
        return self.__cqrs_id

    @property
    def instance_data(self):
        return self.__instance_data

    @property
    def pk(self):
        return self.__instance_pk

    @property
    def queue(self):
        return self.__queue

    def to_dict(self):
        return {
            'signal_type': self.__signal_type,
            'cqrs_id': self.__cqrs_id,
            'instance_data': self.__instance_data,
            'instance_pk': self.__instance_pk,
        }



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import datetime
import sys
import time

from django.core.exceptions import FieldError
from django.core.management.base import BaseCommand, CommandError

import ujson

from dj_cqrs.management.commands.utils import batch_qs
from dj_cqrs.registries import MasterRegistry


DEFAULT_BATCH = 10000
DEFAULT_PROGRESS = False


class Command(BaseCommand):
    help = 'Filter synchronization of certain CQRS model rows over transport to replicas.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--cqrs-id', '-cid',
            help='CQRS_ID of the master model',
            type=str,
            required=True,
        )
        parser.add_argument(
            '--filter', '-f',
            help='Filter kwargs',
            type=str,
            default=None,
        )
        parser.add_argument(
            '--queue', '-q',
            help='Name of the specific replica queue',
            type=str,
            default=None,
        )
        parser.add_argument(
            '--batch', '-b',
            help='Batch size',
            type=int,
            default=DEFAULT_BATCH,
        )
        parser.add_argument(
            '--progress', '-p',
            help='Display progress',
            action='store_true',
        )

    def handle(self, *args, **options):
        model = self._get_model(options)
        progress = self._get_progress(options)
        batch_size = self._get_batch_size(options)

        qs = model._default_manager.none()
        if options['filter']:
            try:
                kwargs = ujson.loads(options['filter'])
                if not isinstance(kwargs, dict):
                    raise ValueError
            except ValueError:
                raise CommandError('Bad filter kwargs!')

            try:
                qs = model._default_manager.filter(**kwargs).order_by()
            except FieldError as e:
                raise CommandError('Bad filter kwargs! {}'.format(str(e)))

        db_count = qs.count()
        if db_count == 0:
            print('No objects found for filter!')
            return

        counter, success_counter = 0, 0
        if progress:
            print('Processing {} records with batch size {}'.format(db_count, batch_size))

        for qs in batch_qs(model.relate_cqrs_serialization(qs), batch_size=batch_size):
            ts = time.time()
            cs = counter
            for instance in qs:
                counter += 1
                try:
                    instance.cqrs_sync(queue=options['queue'])
                    success_counter += 1
                except Exception as e:
                    print('\nSync record failed for pk={}: {}: {}'.format(
                        instance.pk, type(e).__name__, str(e),
                    ))

            if progress:
                rate = (counter - cs) / (time.time() - ts)
                percent = 100 * counter / db_count
                eta = datetime.timedelta(seconds=int((db_count - counter) / rate))
                sys.stdout.write(
                    '\r{} of {} processed - {}% with rate {:.1f} rps, to go {} ...{:20}'.format(
                        counter, db_count, int(percent), rate, str(eta), ' ',
                    ))
                sys.stdout.flush()

        print('Done!\n{} instance(s) synced.\n{} instance(s) processed.'.format(
            success_counter, counter,
        ))

    @staticmethod
    def _get_model(options):
        cqrs_id = options['cqrs_id']
        model = MasterRegistry.get_model_by_cqrs_id(cqrs_id)

        if not model:
            raise CommandError('Wrong CQRS ID: {}!'.format(cqrs_id))

        return model

    @staticmethod
    def _get_batch_size(options):
        return options.get('batch', DEFAULT_BATCH)

    @staticmethod
    def _get_progress(options):
        return bool(options.get('progress', DEFAULT_PROGRESS))



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import ujson

import pytest
from django.core.management import CommandError, call_command
from django.db import transaction
from tests.utils import db_error

from tests.dj_master.models import Author, Publisher
from tests.test_commands.utils import remove_file

COMMAND_NAME = 'cqrs_bulk_dump'


def test_no_cqrs_id():
    with pytest.raises(CommandError):
        call_command(COMMAND_NAME)


def test_bad_cqrs_id():
    with pytest.raises(CommandError) as e:
        call_command(COMMAND_NAME, '--cqrs-id=invalid')

    assert 'Wrong CQRS ID: invalid!' in str(e)


def test_output_file_exists():
    with pytest.raises(CommandError) as e:
        call_command(COMMAND_NAME, '--cqrs-id=author', '-o=setup.py')

    assert 'File setup.py exists!' in str(e)


@pytest.mark.django_db
def test_dumps_no_rows(capsys):
    remove_file('author.dump')

    call_command(COMMAND_NAME, '--cqrs-id=author')

    with open('author.dump', 'r') as f:
        lines = f.readlines()
        assert len(lines) == 1
        assert lines[0] == 'author'

    captured = capsys.readouterr()
    assert 'Done!\n0 instance(s) saved.\n0 instance(s) processed.' in captured.err


@pytest.mark.django_db
def tests_dumps_several_rows(capsys):
    remove_file('author.dump')

    Author.objects.create(id=2, name='2')

    with transaction.atomic():
        publisher = Publisher.objects.create(id=1, name='publisher')
        author = Author.objects.create(id=1, name='1', publisher=publisher)

    call_command(COMMAND_NAME, '--cqrs-id=author')

    with open('author.dump', 'r') as f:
        lines = f.readlines()
        assert len(lines) == 3
        assert lines[0].strip() == 'author'

        line_with_publisher = next(ln for ln in lines[1:] if '"name":"publisher"' in ln)
        assert author.to_cqrs_dict() == ujson.loads(line_with_publisher)

    captured = capsys.readouterr()
    assert 'Done!\n2 instance(s) saved.\n2 instance(s) processed.' in captured.err


@pytest.mark.django_db
def tests_dumps_more_than_batch(capsys):
    remove_file('author.dump')

    Author.objects.bulk_create(
        (Author(id=index, name='n') for index in range(1, 150)),
    )

    call_command(COMMAND_NAME, '--cqrs-id=author')

    with open('author.dump', 'r') as f:
        lines = f.readlines()
        assert len(lines) == 150

    captured = capsys.readouterr()
    assert 'Done!\n149 instance(s) saved.\n149 instance(s) processed.' in captured.err


@pytest.mark.django_db
def test_error(capsys, mocker):
    remove_file('author.dump')
    Author.objects.create(id=2, name='2')

    mocker.patch('tests.dj_master.models.Author.to_cqrs_dict', side_effect=db_error)
    call_command(COMMAND_NAME, '--cqrs-id=author')

    captured = capsys.readouterr()
    assert 'Dump record failed for pk=2' in captured.err
    assert '1 instance(s) processed.' in captured.err
    assert '0 instance(s) saved.' in captured.err


@pytest.mark.django_db
def test_progress(capsys):
    remove_file('author.dump')

    Author.objects.create(id=2, name='2')
    call_command(COMMAND_NAME, '--cqrs-id=author', '--progress')

    captured = capsys.readouterr()
    assert 'Processing 1 records with batch size 10000' in captured.err
    assert '1 of 1 processed - 100% with rate' in captured.err



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

SECRET_KEY = '6^1@%4#qtc+bookwp4w5k-+nbo+clm!skzdhnyl@rf&06b5tl7'

DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',

    'dj_cqrs',
    'tests.dj_master',
    'tests.dj_replica',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'tests.dj.urls'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

CQRS = {
    'transport': 'tests.dj.transport.TransportStub',
    'queue': 'replica',
}



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import sys
from io import StringIO

import pytest
from django.conf import settings
from django.core.management import CommandError, call_command
from django.utils.timezone import now

from tests.dj_master.models import Author
from tests.dj_replica.models import AuthorRef


COMMAND_NAME = 'cqrs_diff_replica'


def test_bad_cqrs_id(mocker):
    mocker.patch.object(sys, 'stdin', StringIO('invalid,datetime\n'))

    with pytest.raises(CommandError) as e:
        call_command(COMMAND_NAME)

    assert 'Wrong CQRS ID: invalid!' in str(e)


@pytest.mark.django_db
def test_no_rows(mocker, capsys):
    mocker.patch.object(sys, 'stdin', StringIO('author,datetime\n'))
    call_command(COMMAND_NAME)

    captured = capsys.readouterr()
    assert not captured.err
    assert 'author,datetime,replica\n' == captured.out


def master_replica_pipe(capsys, mocker, *args):
    call_command('cqrs_diff_master', '--cqrs-id=author', *args)
    captured = capsys.readouterr()
    mocker.patch.object(sys, 'stdin', StringIO(captured.out))
    call_command(COMMAND_NAME)
    return capsys.readouterr()


@pytest.mark.django_db
def test_first_row(capsys, mocker):
    Author.objects.create(name='author', id=1)

    captured = master_replica_pipe(capsys, mocker)

    first_row = captured.out.split('\n')[0]
    assert '{},'.format(Author.CQRS_ID) in first_row
    assert settings.CQRS['queue'] in first_row


@pytest.mark.django_db
def test_all_synced(capsys, mocker):
    mocker.patch('dj_cqrs.controller.producer.produce')

    Author.objects.create(id=1, name='name')
    AuthorRef.objects.create(id=1, name='name', cqrs_revision=0, cqrs_updated=now())

    captured = master_replica_pipe(capsys, mocker)
    assert not captured.err


@pytest.mark.django_db
def test_sync_for_all(mocker, capsys):
    for i in range(3):
        Author.objects.create(name=str(i), id=i)

    captured = master_replica_pipe(capsys, mocker)
    for err_text in ('PK to resync', '0', '1', '2'):
        assert err_text in captured.err

    second_row = captured.out.split('\n')[1]
    assert '[0,1,2]' in second_row


@pytest.mark.django_db
def test_partial_sync(mocker, capsys):
    mocker.patch('dj_cqrs.controller.producer.produce')

    Author.objects.create(id=1, name='name')
    AuthorRef.objects.create(id=1, name='name', cqrs_revision=0, cqrs_updated=now())
    Author.objects.create(id=2, name='2')

    captured = master_replica_pipe(capsys, mocker)
    assert 'PK to resync' in captured.err
    assert '2' in captured.err

    second_row = captured.out.split('\n')[1]
    assert '[2]' in second_row


@pytest.mark.django_db
def test_sync_batch(mocker, capsys):
    mocker.patch('dj_cqrs.controller.producer.produce')

    for i in range(3):
        Author.objects.create(id=i, name=str(i))
    AuthorRef.objects.create(id=1, name='name', cqrs_revision=0, cqrs_updated=now())

    captured = master_replica_pipe(capsys, mocker, '--batch=1')
    for err_text in ('PK to resync', '0', '2'):
        assert err_text in captured.err

    second_row = captured.out.split('\n')[1]
    assert '[0]' in second_row

    third_row = captured.out.split('\n')[2]
    assert '[2]' in third_row



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import logging
import ujson
from importlib import import_module

import pytest
from django.db import DatabaseError
from pika.exceptions import AMQPError
from six.moves import reload_module

from dj_cqrs.constants import SignalType
from dj_cqrs.dataclasses import TransportPayload
from dj_cqrs.transport.rabbit_mq import RabbitMQTransport
from tests.utils import db_error


class PublicRabbitMQTransport(RabbitMQTransport):
    @classmethod
    def get_common_settings(cls):
        return cls._get_common_settings()

    @classmethod
    def get_consumer_settings(cls):
        return cls._get_consumer_settings()

    @classmethod
    def consume_message(cls, *args):
        return cls._consume_message(*args)

    @classmethod
    def produce_message(cls, *args):
        return cls._produce_message(*args)


def test_default_settings():
    s = PublicRabbitMQTransport.get_common_settings()
    assert s[0] == 'localhost'
    assert s[1] == 5672
    assert s[2].username == 'guest' and s[2].password == 'guest'
    assert s[3] == 'cqrs'


def test_non_default_settings(settings):
    settings.CQRS = {
        'transport': 'dj_cqrs.transport.rabbit_mq.RabbitMQTransport',
        'host': 'rabbit',
        'port': 8000,
        'user': 'usr',
        'password': 'pswd',
        'exchange': 'exchange',
    }

    s = PublicRabbitMQTransport.get_common_settings()
    assert s[0] == 'rabbit'
    assert s[1] == 8000
    assert s[2].username == 'usr' and s[2].password == 'pswd'
    assert s[3] == 'exchange'


def test_consumer_default_settings():
    s = PublicRabbitMQTransport.get_consumer_settings()
    assert s[0] is None
    assert s[1] == 10


def test_consumer_non_default_settings(settings):
    settings.CQRS = {
        'transport': 'dj_cqrs.transport.rabbit_mq.RabbitMQTransport',
        'queue': 'q',
        'consumer_prefetch_count': 2,
    }

    s = PublicRabbitMQTransport.get_consumer_settings()
    assert s[0] == 'q'
    assert s[1] == 2


@pytest.fixture
def rabbit_transport(settings):
    settings.CQRS = {
        'transport': 'dj_cqrs.transport.rabbit_mq.RabbitMQTransport',
        'queue': 'replica',
    }
    module = reload_module(import_module('dj_cqrs.transport'))
    yield module.current_transport


def amqp_error(*args, **kwargs):
    raise AMQPError


def test_produce_connection_error(rabbit_transport, mocker, caplog):
    mocker.patch.object(RabbitMQTransport, '_get_producer_rmq_objects', side_effect=amqp_error)

    rabbit_transport.produce(
        TransportPayload(
            SignalType.SAVE, 'CQRS_ID', {'id': 1}, 1,
        ),
    )
    assert "CQRS couldn't be published: pk = 1 (CQRS_ID)." in caplog.text


def test_produce_publish_error(rabbit_transport, mocker, caplog):
    mocker.patch.object(
        RabbitMQTransport, '_get_producer_rmq_objects', return_value=(mocker.MagicMock(), None),
    )
    mocker.patch.object(RabbitMQTransport, '_produce_message', side_effect=amqp_error)

    rabbit_transport.produce(
        TransportPayload(
            SignalType.SAVE, 'CQRS_ID', {'id': 1}, 1,
        ),
    )
    assert "CQRS couldn't be published: pk = 1 (CQRS_ID)." in caplog.text


def test_produce_ok(rabbit_transport, mocker, caplog):
    caplog.set_level(logging.INFO)
    mocker.patch.object(
        RabbitMQTransport, '_get_producer_rmq_objects', return_value=(mocker.MagicMock(), None),
    )
    mocker.patch.object(RabbitMQTransport, '_produce_message', return_value=True)

    rabbit_transport.produce(
        TransportPayload(
            SignalType.SAVE, 'CQRS_ID', {'id': 1}, 1,
        ),
    )
    assert 'CQRS is published: pk = 1 (CQRS_ID).' in caplog.text


def test_produce_message_ok(mocker):
    channel = mocker.MagicMock()
    payload = TransportPayload(SignalType.SAVE, 'cqrs_id', {}, 'id')

    PublicRabbitMQTransport.produce_message(channel, 'exchange', payload)

    assert channel.basic_publish.call_count == 1

    basic_publish_kwargs = channel.basic_publish.call_args[1]
    assert ujson.loads(basic_publish_kwargs['body']) == \
        {
            'signal_type': SignalType.SAVE,
            'cqrs_id': 'cqrs_id',
            'instance_data': {},
            'instance_pk': 'id',
        }
    assert basic_publish_kwargs['exchange'] == 'exchange'
    assert basic_publish_kwargs['mandatory']
    assert basic_publish_kwargs['routing_key'] == 'cqrs_id'
    assert basic_publish_kwargs['properties'].content_type == 'text/plain'
    assert basic_publish_kwargs['properties'].delivery_mode == 2


def test_produce_sync_message_no_queue(mocker):
    channel = mocker.MagicMock()
    payload = TransportPayload(SignalType.SYNC, 'cqrs_id', {}, None)

    PublicRabbitMQTransport.produce_message(channel, 'exchange', payload)

    basic_publish_kwargs = channel.basic_publish.call_args[1]
    assert ujson.loads(basic_publish_kwargs['body']) == \
        {
            'signal_type': SignalType.SYNC,
            'cqrs_id': 'cqrs_id',
            'instance_data': {},
            'instance_pk': None,
        }
    assert basic_publish_kwargs['routing_key'] == 'cqrs_id'


def test_produce_sync_message_queue(mocker):
    channel = mocker.MagicMock()
    payload = TransportPayload(SignalType.SYNC, 'cqrs_id', {}, 'id', 'queue')

    PublicRabbitMQTransport.produce_message(channel, 'exchange', payload)

    basic_publish_kwargs = channel.basic_publish.call_args[1]
    assert ujson.loads(basic_publish_kwargs['body']) == \
        {
            'signal_type': SignalType.SYNC,
            'cqrs_id': 'cqrs_id',
            'instance_data': {},
            'instance_pk': 'id',
        }
    assert basic_publish_kwargs['routing_key'] == 'cqrs.queue.cqrs_id'


def test_consume_connection_error(rabbit_transport, mocker, caplog):
    mocker.patch.object(
        RabbitMQTransport, '_get_consumer_rmq_objects', side_effect=amqp_error,
    )
    mocker.patch('time.sleep', side_effect=db_error)

    with pytest.raises(DatabaseError):
        rabbit_transport.consume()

    assert 'AMQP connection error. Reconnecting...' in caplog.text


def test_consume_ok(rabbit_transport, mocker):
    channel_mock = mocker.MagicMock()
    channel_mock.start_consuming = db_error

    mocker.patch.object(
        RabbitMQTransport, '_get_consumer_rmq_objects', return_value=(None, channel_mock),
    )

    with pytest.raises(DatabaseError):
        rabbit_transport.consume()


def test_consume_message_ack(mocker, caplog):
    caplog.set_level(logging.INFO)
    consumer_mock = mocker.patch('dj_cqrs.controller.consumer.consume')

    PublicRabbitMQTransport.consume_message(
        mocker.MagicMock(),
        mocker.MagicMock(),
        None,
        '{"signal_type":"signal","cqrs_id":"cqrs_id","instance_data":{},"instance_pk":1}',
    )

    assert consumer_mock.call_count == 1

    payload = consumer_mock.call_args[0][0]
    assert payload.signal_type == 'signal'
    assert payload.cqrs_id == 'cqrs_id'
    assert payload.instance_data == {}
    assert payload.pk == 1

    assert 'CQRS is received: pk = 1 (cqrs_id).' in caplog.text
    assert 'CQRS is applied: pk = 1 (cqrs_id).' in caplog.text


def test_consume_message_ack_deprecated_structure(mocker, caplog):
    caplog.set_level(logging.INFO)
    consumer_mock = mocker.patch('dj_cqrs.controller.consumer.consume')

    PublicRabbitMQTransport.consume_message(
        mocker.MagicMock(),
        mocker.MagicMock(),
        None,
        '{"signal_type":"signal","cqrs_id":"cqrs_id","instance_data":{}}',
    )

    assert consumer_mock.call_count == 1

    payload = consumer_mock.call_args[0][0]
    assert payload.signal_type == 'signal'
    assert payload.cqrs_id == 'cqrs_id'
    assert payload.instance_data == {}
    assert payload.pk is None

    assert 'CQRS deprecated package structure.' in caplog.text
    assert 'CQRS is received: pk = None (cqrs_id).' not in caplog.text
    assert 'CQRS is applied: pk = None (cqrs_id).' not in caplog.text


def test_consume_message_nack(mocker, caplog):
    caplog.set_level(logging.INFO)
    mocker.patch('dj_cqrs.controller.consumer.consume', return_value=None)

    PublicRabbitMQTransport.consume_message(
        mocker.MagicMock(),
        mocker.MagicMock(),
        None,
        '{"signal_type":"signal","cqrs_id":"cqrs_id","instance_data":{},"instance_pk":1}',
    )

    assert 'CQRS is received: pk = 1 (cqrs_id).' in caplog.text
    assert 'CQRS is denied: pk = 1 (cqrs_id).' in caplog.text


def test_consume_message_nack_deprecated_structure(mocker, caplog):
    caplog.set_level(logging.INFO)
    mocker.patch('dj_cqrs.controller.consumer.consume', return_value=None)

    PublicRabbitMQTransport.consume_message(
        mocker.MagicMock(),
        mocker.MagicMock(),
        None,
        '{"signal_type":"signal","cqrs_id":"cqrs_id","instance_data":{}}',
    )

    assert 'CQRS is received: pk = 1 (cqrs_id).' not in caplog.text
    assert 'CQRS is denied: pk = 1 (cqrs_id).' not in caplog.text


def test_consume_message_json_parsing_error(mocker, caplog):
    PublicRabbitMQTransport.consume_message(
        mocker.MagicMock(), mocker.MagicMock(), None, '{bad_payload:',
    )

    assert "CQRS couldn't be parsed: {bad_payload:." in caplog.text


def test_consume_message_package_structure_error(mocker, caplog):
    PublicRabbitMQTransport.consume_message(
        mocker.MagicMock(), mocker.MagicMock(), None, '{"pk":"1"}',
    )

    assert """CQRS couldn't be parsed: {"pk":"1"}""" in caplog.text



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import os

from django.core.wsgi import get_wsgi_application

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'integration_tests.replica_settings')

application = get_wsgi_application()



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import pytest
from django.db.models.signals import post_delete, post_save

from dj_cqrs.signals import post_bulk_create, post_update
from dj_cqrs.constants import SignalType
from tests.dj_master import models
from tests.utils import assert_publisher_once_called_with_args


@pytest.mark.parametrize('model', (models.AllFieldsModel, models.BasicFieldsModel))
@pytest.mark.parametrize('signal', (post_delete, post_save, post_bulk_create, post_update))
def test_signals_are_registered(model, signal):
    assert signal.has_listeners(model)


@pytest.mark.django_db(transaction=True)
def test_post_save_create(mocker):
    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')
    models.SimplestModel.objects.create(id=1)

    assert_publisher_once_called_with_args(
        publisher_mock,
        SignalType.SAVE, models.SimplestModel.CQRS_ID, {'id': 1, 'name': None}, 1,
    )


@pytest.mark.django_db(transaction=True)
def test_post_save_update(mocker):
    m = models.SimplestModel.objects.create(id=1)

    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')
    m.name = 'new'
    m.save(update_fields=['name'])

    assert_publisher_once_called_with_args(
        publisher_mock,
        SignalType.SAVE, models.SimplestModel.CQRS_ID, {'id': 1, 'name': 'new'}, 1,
    )


@pytest.mark.django_db(transaction=True)
def test_post_save_delete(mocker):
    m = models.SimplestModel.objects.create(id=1)

    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')
    m.delete()

    assert_publisher_once_called_with_args(
        publisher_mock,
        SignalType.DELETE, models.SimplestModel.CQRS_ID, {'id': 1, 'cqrs_revision': 1}, 1,
    )

    cqrs_updated = publisher_mock.call_args[0][0].to_dict()['instance_data']['cqrs_updated']
    assert isinstance(cqrs_updated, str)


@pytest.mark.django_db(transaction=True)
def test_post_bulk_create(mocker):
    models.AutoFieldsModel.objects.bulk_create([models.AutoFieldsModel() for _ in range(3)])
    created_models = list(models.AutoFieldsModel.objects.all())

    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')
    models.AutoFieldsModel.call_post_bulk_create(created_models)

    assert publisher_mock.call_count == 3


@pytest.mark.django_db(transaction=True)
def test_post_bulk_update(mocker):
    for i in range(3):
        models.SimplestModel.objects.create(id=i)
    cqrs_updated = models.SimplestModel.objects.get(id=1).cqrs_updated

    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')
    models.SimplestModel.cqrs.bulk_update(
        queryset=models.SimplestModel.objects.filter(id__in={1}),
        name='new',
    )

    assert_publisher_once_called_with_args(
        publisher_mock,
        SignalType.SAVE, models.SimplestModel.CQRS_ID, {'id': 1, 'name': 'new'}, 1,
    )

    m = models.SimplestModel.objects.get(id=1)
    assert m.cqrs_updated > cqrs_updated
    assert m.cqrs_revision == 1



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import pytest

from integration_tests.tests.utils import (
    REPLICA_BASIC_TABLE, REPLICA_EVENT_TABLE,
    count_replica_rows, get_replica_all, transport_delay,
)
from tests.dj_master.models import BasicFieldsModel


@pytest.mark.django_db(transaction=True)
def test_both_consumers_consume(replica_cursor):
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 0
    assert count_replica_rows(replica_cursor, REPLICA_EVENT_TABLE) == 0

    master_instances = BasicFieldsModel.objects.bulk_create([
        BasicFieldsModel(
            int_field=index,
            char_field='text',
        )
        for index in range(1, 10)
    ])
    BasicFieldsModel.call_post_bulk_create(master_instances)

    transport_delay(3)
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 9
    assert count_replica_rows(replica_cursor, REPLICA_EVENT_TABLE) == 9

    events_data = get_replica_all(replica_cursor, REPLICA_EVENT_TABLE, ('pid', ),)
    assert len({d[0] for d in events_data}) == 2


@pytest.mark.django_db(transaction=True)
def test_de_duplication(replica_cursor):
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 0
    assert count_replica_rows(replica_cursor, REPLICA_EVENT_TABLE) == 0

    master_instance = BasicFieldsModel.objects.create(int_field=1, char_field='text')
    BasicFieldsModel.call_post_bulk_create([master_instance for _ in range(9)])

    transport_delay(3)
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 1
    assert count_replica_rows(replica_cursor, REPLICA_EVENT_TABLE) == 10



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

from django.db import models

from dj_cqrs.mixins import ReplicaMixin


class BasicFieldsModelRef(ReplicaMixin, models.Model):
    CQRS_ID = 'basic'

    int_field = models.IntegerField(primary_key=True)
    char_field = models.CharField(max_length=200)

    bool_field = models.NullBooleanField()
    date_field = models.DateField(null=True)
    datetime_field = models.DateTimeField(null=True)
    float_field = models.FloatField(null=True)
    url_field = models.URLField(null=True)
    uuid_field = models.UUIDField(null=True)


class BadTypeModelRef(ReplicaMixin, models.Model):
    CQRS_ID = 'basic_1'

    int_field = models.IntegerField(primary_key=True)
    datetime_field = models.NullBooleanField()


class MappedFieldsModelRef(ReplicaMixin, models.Model):
    CQRS_ID = 'basic_2'
    CQRS_MAPPING = {
        'int_field': 'id',
        'char_field': 'name',
    }

    id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=200)


class BadMappingModelRef(ReplicaMixin, models.Model):
    CQRS_ID = 'basic_3'
    CQRS_MAPPING = {
        'int_field': 'id',
        'invalid_field': 'name',
    }

    id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=200)


class Event(models.Model):
    pid = models.IntegerField()
    cqrs_id = models.CharField(max_length=20)
    cqrs_revision = models.IntegerField()

    time = models.DateTimeField(auto_now_add=True)


class Publisher(models.Model):
    id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=20)


class AuthorRef(ReplicaMixin, models.Model):
    CQRS_ID = 'author'
    CQRS_CUSTOM_SERIALIZATION = True

    id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=20)

    publisher = models.ForeignKey(Publisher, null=True, on_delete=models.CASCADE)

    @classmethod
    def cqrs_create(cls, sync, **mapped_data):
        publisher_data, publisher = mapped_data.pop('publisher', None), None
        if publisher_data:
            publisher, _ = Publisher.objects.get_or_create(**publisher_data)

        books_data = mapped_data.pop('books', [])
        author = cls.objects.create(publisher=publisher, **mapped_data)

        Book.objects.bulk_create(Book(author=author, **book_data) for book_data in books_data)
        return author

    def cqrs_update(self, sync, **mapped_data):
        # It's just an example, that doesn't make sense in real cases
        publisher_data, publisher = mapped_data.pop('publisher', None), None
        if publisher_data:
            publisher, _ = Publisher.objects.get_or_create(**publisher_data)

        self.publisher = publisher
        self.name = mapped_data.get('name', self.name)
        self.cqrs_revision = mapped_data['cqrs_revision']
        self.cqrs_updated = mapped_data['cqrs_updated']
        self.save()

        return self


class Book(models.Model):
    id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=20)

    author = models.ForeignKey(AuthorRef, on_delete=models.CASCADE)



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import sys
from io import StringIO

import pytest
from django.core.management import CommandError, call_command
from django.utils.timezone import now

from dj_cqrs.constants import NO_QUEUE
from dj_cqrs.management.commands import cqrs_sync
from tests.dj_master.models import Author
from tests.dj_replica.models import AuthorRef


COMMAND_NAME = 'cqrs_diff_sync'


def test_bad_cqrs_id(mocker):
    mocker.patch.object(sys, 'stdin', StringIO('invalid,datetime,replica\n'))

    with pytest.raises(CommandError) as e:
        call_command(COMMAND_NAME)

    assert 'Wrong CQRS ID: invalid!' in str(e)


def diff_pipe(capsys, mocker, *args):
    call_command('cqrs_diff_master', '--cqrs-id=author', *args)
    captured = capsys.readouterr()

    mocker.patch.object(sys, 'stdin', StringIO(captured.out))
    call_command('cqrs_diff_replica')
    captured = capsys.readouterr()
    mocker.stopall()

    sync_mock = mocker.patch.object(cqrs_sync.Command, 'handle')
    mocker.patch.object(sys, 'stdin', StringIO(captured.out))
    call_command(COMMAND_NAME)

    return sync_mock


@pytest.mark.django_db
def test_all_synced(capsys, mocker):
    mocker.patch('dj_cqrs.controller.producer.produce')

    Author.objects.create(id=1, name='name')
    AuthorRef.objects.create(id=1, name='name', cqrs_revision=0, cqrs_updated=now())

    sync_mock = diff_pipe(capsys, mocker)
    sync_mock.assert_not_called()


@pytest.mark.django_db
def test_sync_for_all(mocker, capsys):
    mocker.patch('dj_cqrs.controller.producer.produce')
    Author.objects.create(name='a', id=1)

    sync_mock = diff_pipe(capsys, mocker)
    sync_mock.assert_called_once_with(
        **{'cqrs_id': 'author', 'filter': '{"id__in": [1]}', 'queue': 'replica'},
    )


@pytest.mark.django_db
def test_partial_sync(mocker, capsys):
    mocker.patch('dj_cqrs.controller.producer.produce')

    Author.objects.create(id=1, name='name')
    AuthorRef.objects.create(id=1, name='name', cqrs_revision=0, cqrs_updated=now())
    Author.objects.create(id=2, name='2')

    sync_mock = diff_pipe(capsys, mocker)
    sync_mock.assert_called_once_with(
        **{'cqrs_id': 'author', 'filter': '{"id__in": [2]}', 'queue': 'replica'},
    )


@pytest.mark.django_db
def test_sync_batch(mocker, capsys):
    mocker.patch('dj_cqrs.controller.producer.produce')

    for i in range(3):
        Author.objects.create(id=i, name=str(i))
    AuthorRef.objects.create(id=1, name='name', cqrs_revision=0, cqrs_updated=now())

    sync_mock = diff_pipe(capsys, mocker, '--batch=1')
    assert sync_mock.call_count == 2


@pytest.mark.django_db
def test_sync_no_queue(mocker):
    sync_mock = mocker.patch.object(cqrs_sync.Command, 'handle')
    mocker.patch.object(sys, 'stdin', StringIO('author,dt,{}\n[1]\n'.format(NO_QUEUE)))
    call_command(COMMAND_NAME)

    sync_mock.assert_called_once_with(**{'cqrs_id': 'author', 'filter': '{"id__in": [1]}'})



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import random

from django.core.management.base import BaseCommand
from django.db import transaction

from tests.dj_master.models import Author, Book, Publisher


class Command(BaseCommand):
    help = 'Simulate N signals.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--count', '-c', help='Simulation of N signals.', type=int, default=3000,
        )

    @staticmethod
    def _get_max_id(model):
        try:
            return model.objects.all().order_by('-id')[0].id
        except IndexError:
            return 0

    def handle(self, *args, **options):
        max_author_id = self._get_max_id(Author)
        max_book_id = self._get_max_id(Book)
        max_publisher_id = self._get_max_id(Publisher)

        with transaction.atomic():
            for _ in range(options['count']):
                publisher = None
                if bool(random.getrandbits(1)):
                    max_publisher_id += 1
                    publisher = Publisher.objects.create(id=max_publisher_id, name='p')

                max_author_id += 1
                author = Author.objects.create(id=max_author_id, name='a', publisher=publisher)

                for _ in range(random.randint(0, 2)):
                    max_book_id += 1
                    Book.objects.create(id=max_book_id, title='t', author=author)



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import pytest
from uuid import uuid4

from django.db import transaction
from django.db.models import CharField, IntegerField
from django.utils.timezone import now

from dj_cqrs.constants import SignalType
from dj_cqrs.metas import MasterMeta
from tests.dj_master import models
from tests.dj_master.serializers import AuthorSerializer
from tests.utils import assert_is_sub_dict, assert_publisher_once_called_with_args


class MasterMetaTest(MasterMeta):
    @classmethod
    def check_cqrs_fields(cls, model_cls):
        return cls._check_cqrs_fields(model_cls)

    @classmethod
    def check_correct_configuration(cls, model_cls):
        return cls._check_correct_configuration(model_cls)


def test_cqrs_fields_non_existing_field(mocker):
    with pytest.raises(AssertionError) as e:

        class Cls(object):
            CQRD_ID = 'ID'
            CQRS_FIELDS = ('char_field', 'integer_field')

            char_field = CharField(max_length=100, primary_key=True)
            int_field = IntegerField()

            _meta = mocker.MagicMock(concrete_fields=(char_field, int_field), private_fields=())
            _meta.pk.name = 'char_field'

        MasterMetaTest.check_cqrs_fields(Cls)

    assert str(e.value) == 'CQRS_FIELDS field is not correctly set for model Cls.'


def test_cqrs_fields_id_is_not_included(mocker):
    with pytest.raises(AssertionError) as e:

        class Cls(object):
            CQRD_ID = 'ID'
            CQRS_FIELDS = ('int_field',)

            char_field = CharField(max_length=100, primary_key=True)
            int_field = IntegerField()

            _meta = mocker.MagicMock(concrete_fields=(char_field, int_field), private_fields=())
            _meta.pk.name = 'char_field'

        MasterMetaTest.check_cqrs_fields(Cls)

    assert str(e.value) == 'PK is not in CQRS_FIELDS for model Cls.'


def test_cqrs_fields_duplicates(mocker):
    with pytest.raises(AssertionError) as e:

        class Cls(object):
            CQRD_ID = 'ID'
            CQRS_FIELDS = ('char_field', 'char_field')

            char_field = CharField(max_length=100, primary_key=True)
            int_field = IntegerField()

            _meta = mocker.MagicMock(concrete_fields=(char_field, int_field), private_fields=())
            _meta.pk.name = 'char_field'

        MasterMetaTest.check_cqrs_fields(Cls)

    assert str(e.value) == 'Duplicate names in CQRS_FIELDS field for model Cls.'


def test_cqrs_bad_configuration(mocker):
    with pytest.raises(AssertionError) as e:

        class Cls(object):
            CQRD_ID = 'ID'
            CQRS_FIELDS = ('char_field',)
            CQRS_SERIALIZER = 'path.to.serializer'

            char_field = CharField(max_length=100, primary_key=True)

            _meta = mocker.MagicMock(concrete_fields=(char_field,), private_fields=())
            _meta.pk.name = 'char_field'

        MasterMetaTest.check_correct_configuration(Cls)

    assert "CQRS_FIELDS can't be set together with CQRS_SERIALIZER." in str(e.value)


@pytest.mark.django_db
def test_to_cqrs_dict_has_cqrs_fields():
    m = models.AutoFieldsModel.objects.create()
    dct = m.to_cqrs_dict()
    assert dct['cqrs_revision'] == 0 and dct['cqrs_updated'] is not None


def test_to_cqrs_dict_basic_types():
    dt = now()
    uid = uuid4()
    m = models.BasicFieldsModel(
        int_field=1,
        bool_field=False,
        char_field='str',
        date_field=None,
        datetime_field=dt,
        float_field=1.23,
        url_field='http://example.com',
        uuid_field=uid,
    )
    assert_is_sub_dict({
        'int_field': 1,
        'bool_field': False,
        'char_field': 'str',
        'date_field': None,
        'datetime_field': str(dt),
        'float_field': 1.23,
        'url_field': 'http://example.com',
        'uuid_field': uid,
    }, m.to_cqrs_dict())


def test_to_cqrs_dict_all_fields():
    m = models.AllFieldsModel(char_field='str')
    assert_is_sub_dict({'id': None, 'int_field': None, 'char_field': 'str'}, m.to_cqrs_dict())


def test_to_cqrs_dict_chosen_fields():
    m = models.ChosenFieldsModel(float_field=1.23)
    assert_is_sub_dict({'char_field': None, 'id': None}, m.to_cqrs_dict())


@pytest.mark.django_db
def test_to_cqrs_dict_auto_fields():
    m = models.AutoFieldsModel()
    assert_is_sub_dict({'id': None, 'created': None, 'updated': None}, m.to_cqrs_dict())

    m.save()
    cqrs_dct = m.to_cqrs_dict()
    for key in ('id', 'created', 'updated'):
        assert cqrs_dct[key] is not None


def test_cqrs_sync_not_created():
    m = models.BasicFieldsModel()
    assert not m.cqrs_sync()


@pytest.mark.django_db
def test_cqrs_sync_cant_refresh_model():
    m = models.SimplestModel.objects.create()
    assert not m.cqrs_sync()


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_cqrs_sync_not_saved(mocker):
    m = models.ChosenFieldsModel.objects.create(char_field='old')
    m.name = 'new'

    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')

    assert m.cqrs_sync()

    assert_publisher_once_called_with_args(
        publisher_mock,
        SignalType.SYNC, models.ChosenFieldsModel.CQRS_ID, {'char_field': 'old', 'id': m.pk}, m.pk,
    )


@pytest.mark.django_db(transaction=True)
def test_cqrs_sync(mocker):
    m = models.ChosenFieldsModel.objects.create(char_field='old')
    m.char_field = 'new'
    m.save(update_fields=['char_field'])

    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')

    assert m.cqrs_sync()

    assert_publisher_once_called_with_args(
        publisher_mock,
        SignalType.SYNC, models.ChosenFieldsModel.CQRS_ID, {'char_field': 'new', 'id': m.pk}, m.pk,
    )


@pytest.mark.django_db(transaction=True)
def test_is_sync_instance(mocker):
    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')

    out_instance = models.FilteredSimplestModel.objects.create(name='a')
    in_instance = models.FilteredSimplestModel.objects.create(name='title')

    instances = (out_instance, in_instance)
    for instance in instances:
        assert instance.cqrs_revision == 0

    assert_publisher_once_called_with_args(
        publisher_mock,
        SignalType.SAVE,
        models.FilteredSimplestModel.CQRS_ID,
        {'name': 'title', 'id': in_instance.pk},
        in_instance.pk,
    )
    publisher_mock.reset_mock()

    in_instance.name = 'longer title'
    for instance in instances:
        instance.save()
        instance.refresh_from_db()

        assert instance.cqrs_revision == 1

    assert_publisher_once_called_with_args(
        publisher_mock,
        SignalType.SAVE,
        models.FilteredSimplestModel.CQRS_ID,
        {'name': 'longer title', 'id': in_instance.pk},
        in_instance.pk,
    )
    publisher_mock.reset_mock()

    out_instance.name = 'long'
    in_instance.name = 's'
    for instance in instances:
        instance.save()
        instance.refresh_from_db()

        assert instance.cqrs_revision == 2

    assert_publisher_once_called_with_args(
        publisher_mock,
        SignalType.SAVE,
        models.FilteredSimplestModel.CQRS_ID,
        {'name': 'long', 'id': out_instance.pk},
        out_instance.pk,
    )

    publisher_mock.reset_mock()
    in_instance.delete()
    assert publisher_mock.call_count == 0

    out_instance.delete()
    assert publisher_mock.call_count == 1


@pytest.mark.django_db
def test_create():
    for _ in range(2):
        m = models.AutoFieldsModel.objects.create()
        assert m.cqrs_revision == 0
        assert m.cqrs_updated is not None


@pytest.mark.django_db
def test_update():
    m = models.AutoFieldsModel.objects.create()
    cqrs_updated = m.cqrs_updated

    for i in range(1, 3):
        m.save()
        m.refresh_from_db()

        assert m.cqrs_revision == i

        assert m.cqrs_updated > cqrs_updated
        cqrs_updated = m.cqrs_updated


@pytest.mark.django_db(transaction=True)
def test_transaction_rollbacked(mocker):
    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')

    try:
        with transaction.atomic():
            models.BasicFieldsModel.objects.create(
                int_field=1,
                char_field='str',
            )
            raise ValueError

    except ValueError:
        publisher_mock.assert_not_called()


@pytest.mark.django_db(transaction=True)
def test_transaction_commited(mocker):
    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')

    with transaction.atomic():
        models.BasicFieldsModel.objects.create(
            int_field=1,
            char_field='str',
        )

    assert_publisher_once_called_with_args(
        publisher_mock,
        SignalType.SAVE, models.BasicFieldsModel.CQRS_ID, {'char_field': 'str', 'int_field': 1}, 1,
    )


@pytest.mark.django_db(transaction=True)
def test_transaction_rollbacked_to_savepoint(mocker):
    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')

    with transaction.atomic():
        models.BasicFieldsModel.objects.create(
            int_field=1,
            char_field='str',
        )

        try:
            with transaction.atomic(savepoint=True):
                raise ValueError
        except ValueError:
            pass

    assert_publisher_once_called_with_args(
        publisher_mock,
        SignalType.SAVE, models.BasicFieldsModel.CQRS_ID, {'char_field': 'str', 'int_field': 1}, 1,
    )


@pytest.mark.django_db(transaction=True)
def test_serialization_no_related_instance(mocker):
    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')
    models.Author.objects.create(id=1, name='author', publisher=None)

    assert_publisher_once_called_with_args(
        publisher_mock,
        SignalType.SAVE, models.Author.CQRS_ID,
        {
            'id': 1,
            'name': 'author',
            'publisher': None,
            'books': [],
            'cqrs_revision': 0,
        }, 1,
    )


@pytest.mark.django_db(transaction=True)
def test_save_serialization(mocker, django_assert_num_queries):
    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')

    # 0 - Transaction start
    # 1 - Publisher
    # 2 - Author
    # 3-4 - Books
    # 5-6 - Serialization with prefetch_related
    query_counter = 7
    with django_assert_num_queries(query_counter):
        with transaction.atomic():
            publisher = models.Publisher.objects.create(id=1, name='publisher')
            author = models.Author.objects.create(id=1, name='author', publisher=publisher)
            for index in range(1, 3):
                models.Book.objects.create(id=index, title=str(index), author=author)

    assert_publisher_once_called_with_args(
        publisher_mock,
        SignalType.SAVE, models.Author.CQRS_ID,
        {
            'id': 1,
            'name': 'author',
            'publisher': {
                'id': 1,
                'name': 'publisher',
            },
            'books': [{
                'id': 1,
                'name': '1',
            }, {
                'id': 2,
                'name': '2',
            }],
            'cqrs_revision': 0,
        }, 1,
    )


@pytest.mark.django_db(transaction=True)
def test_delete_serialization():
    m = models.Author.objects.create(id=1, name='author', publisher=None)
    assert models.Author.objects.count() == 1

    m.delete()
    assert models.Author.objects.count() == 0


@pytest.mark.django_db(transaction=True)
def test_create_from_related_table(mocker):
    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')
    author = models.Author.objects.create(id=1, name='author', publisher=None)

    with transaction.atomic():
        models.Book.objects.create(id=1, title='title', author=author)

        # Calling author.cqrs_sync() would result in a wrong cqrs_revision!
        author.save()

    assert publisher_mock.call_count == 2
    assert_is_sub_dict(
        {
            'id': 1,
            'name': 'author',
            'publisher': None,
            'books': [{
                'id': 1,
                'name': 'title',
            }],
            'cqrs_revision': 1,
        },
        publisher_mock.call_args[0][0].instance_data,
    )


@pytest.mark.django_db(transaction=True)
def test_update_from_related_table(mocker):
    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')

    with transaction.atomic():
        publisher = models.Publisher.objects.create(id=1, name='publisher')
        author = models.Author.objects.create(id=1, name='author', publisher=publisher)

    with transaction.atomic():
        publisher.name = 'new'
        publisher.save()

        author.save()

    assert publisher_mock.call_count == 2
    assert_is_sub_dict(
        {
            'id': 1,
            'name': 'author',
            'publisher': {
                'id': 1,
                'name': 'new'
            },
            'books': [],
            'cqrs_revision': 1,
        },
        publisher_mock.call_args[0][0].instance_data,
    )


@pytest.mark.django_db(transaction=True)
def test_to_cqrs_dict_serializer_ok():
    model = models.Author.objects.create(id=1)

    assert_is_sub_dict(AuthorSerializer(model).data, model.to_cqrs_dict())
    assert models.Author._cqrs_serializer_class == AuthorSerializer


@pytest.mark.django_db(transaction=True)
def test_to_cqrs_dict_serializer_import_error():
    with pytest.raises(ImportError) as e:
        with transaction.atomic():
            models.BadSerializationClassModel.objects.create(id=1)

    assert "CQRS_SERIALIZER can't be imported." in str(e)


@pytest.mark.django_db(transaction=True)
def test_to_cqrs_dict_serializer_bad_related_function():
    with pytest.raises(RuntimeError) as e:
        models.BadQuerySetSerializationClassModel.objects.create()

    assert "Couldn't serialize CQRS class (bad_queryset)." in str(e)


@pytest.mark.django_db(transaction=True)
def test_multiple_inheritance(mocker):
    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')

    models.NonMetaClassModel.objects.create(name='abc')

    assert publisher_mock.call_count == 1


@pytest.mark.django_db(transaction=True)
def test_non_sent(mocker):
    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')

    m = models.NonSentModel.objects.create()
    assert publisher_mock.call_count == 0
    m.refresh_from_db()
    assert m.cqrs_revision == 0

    m.save()
    assert publisher_mock.call_count == 0
    m.refresh_from_db()
    assert m.cqrs_revision == 1

    m.delete()
    assert publisher_mock.call_count == 0



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import os
import sys

import ujson
from django.core.management.base import BaseCommand, CommandError
from django.db import transaction, DatabaseError

from dj_cqrs.registries import ReplicaRegistry


class Command(BaseCommand):
    help = 'Bulk load of a CQRS model to a replica service.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--input', '-i',
            help='Input file for loading (- for reading from stdin)',
            type=str, required=True,
        )
        parser.add_argument(
            '--clear', '-c',
            help='Delete existing models',
            type=bool,
            required=False,
            default=False,
        )
        parser.add_argument(
            '--batch', '-b',
            help='Batch size',
            type=int,
            default=10000,
        )

    def handle(self, *args, **options):
        batch_size = self._get_batch_size(options)

        f_name = options['input']
        if f_name != '-' and not os.path.exists(f_name):
            raise CommandError("File {} doesn't exist!".format(f_name))

        with sys.stdin if f_name == '-' else open(f_name, 'r') as f:
            try:
                cqrs_id = next(f).strip()
            except StopIteration:
                cqrs_id = None

            if not cqrs_id:
                raise CommandError('File {} is empty!'.format(f_name))

            model = ReplicaRegistry.get_model_by_cqrs_id(cqrs_id)
            if not model:
                raise CommandError('Wrong CQRS ID: {}!'.format(cqrs_id))

            success_counter = 0
            with transaction.atomic():
                if options['clear']:
                    try:
                        model._default_manager.all().delete()
                    except DatabaseError:
                        raise CommandError("Delete operation fails!")

            line_number = 2
            while True:
                with transaction.atomic():
                    try:
                        for _ in range(0, batch_size):
                            line = f.readline()

                            if not line:
                                raise EOFError
                            try:
                                try:
                                    master_data = ujson.loads(line.strip())
                                except ValueError:
                                    print(
                                        "Dump file can't be parsed: line {}!".format(
                                            line_number,
                                        ),
                                        file=sys.stderr
                                    )
                                    line_number += 1
                                    continue

                                instance = model.cqrs_save(master_data)
                                if not instance:
                                    print(
                                        "Instance can't be saved: line {}!".format(line_number),
                                        file=sys.stderr
                                    )
                                else:
                                    success_counter += 1
                            except Exception as e:
                                print(
                                    'Unexpected error: line {}! {}'.format(line_number, str(e)),
                                    file=sys.stderr
                                )
                            line_number += 1
                    except EOFError:
                        break

            print('Done!\n{} instance(s) loaded.'.format(success_counter), file=sys.stderr)

    @staticmethod
    def _get_batch_size(options):
        return options['batch']



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import pytest

from dj_cqrs.registries import MasterRegistry, ReplicaRegistry
from tests.dj_master import models as master_models
from tests.dj_replica import models as replica_models


@pytest.mark.parametrize('registry', [MasterRegistry, ReplicaRegistry])
def test_duplicate_cqrs_id(registry):
    class Cls(object):
        CQRS_ID = 'basic'

    with pytest.raises(AssertionError) as e:
        registry.register_model(Cls)

    assert str(e.value) == "Two models can't have the same CQRS_ID: basic."


@pytest.mark.parametrize('model,registry', (
    (master_models.SimplestModel, MasterRegistry),
    (master_models.AutoFieldsModel, MasterRegistry),
    (replica_models.BasicFieldsModelRef, ReplicaRegistry),
    (replica_models.BadTypeModelRef, ReplicaRegistry),
))
def test_models_are_registered(model, registry):
    assert registry.models[model.CQRS_ID] == model
    assert registry.get_model_by_cqrs_id(model.CQRS_ID) == model


def test_get_model_by_cqrs_id_no_id(caplog):
    assert ReplicaRegistry.get_model_by_cqrs_id('invalid') is None
    assert 'No model with such CQRS_ID: invalid.' in caplog.text


def test_no_cqrs_queue(settings):
    settings.CQRS.update({'queue': None})

    with pytest.raises(AssertionError) as e:
        ReplicaRegistry.register_model(replica_models.MappedFieldsModelRef)

    assert str(e.value) == 'CQRS queue must be set for the service, that has replica models.'



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import pytest

from dj_cqrs.constants import SignalType
from dj_cqrs.controller.consumer import route_signal_to_replica_model
from dj_cqrs.mixins import ReplicaMixin


def test_bad_model(caplog):
    route_signal_to_replica_model(SignalType.SAVE, 'invalid', {})
    assert 'No model with such CQRS_ID: invalid.' in caplog.text


def test_bad_signal(caplog):
    route_signal_to_replica_model('invalid', 'basic', {})
    assert 'Bad signal type "invalid" for CQRS_ID "basic".' in caplog.text


@pytest.mark.django_db
def test_save_model(mocker):
    cqrs_save_mock = mocker.patch.object(ReplicaMixin, 'cqrs_save')
    route_signal_to_replica_model(SignalType.SAVE, 'basic', {})

    cqrs_save_mock.assert_called_once_with({})


@pytest.mark.django_db
def test_delete_model(mocker):
    cqrs_delete_mock = mocker.patch.object(ReplicaMixin, 'cqrs_delete')
    route_signal_to_replica_model(SignalType.DELETE, 'basic', {'id': 1})

    cqrs_delete_mock.assert_called_once_with({'id': 1})



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

from itertools import chain

import six
from django.db.models import DateField, DateTimeField, F, IntegerField, Manager, Model
from django.db.models.expressions import CombinedExpression
from django.utils.module_loading import import_string

from dj_cqrs.constants import ALL_BASIC_FIELDS
from dj_cqrs.managers import MasterManager, ReplicaManager
from dj_cqrs.metas import MasterMeta, ReplicaMeta
from dj_cqrs.signals import MasterSignals, post_bulk_create, post_update


class RawMasterMixin(Model):
    CQRS_ID = None
    CQRS_PRODUCE = True

    CQRS_FIELDS = ALL_BASIC_FIELDS
    CQRS_SERIALIZER = None  # It must be None or string, like: 'path.to.serializer'

    objects = Manager()
    cqrs = MasterManager()

    cqrs_revision = IntegerField(
        default=0, help_text="This field must be incremented on any model update. "
                             "It's used to for CQRS sync.",
    )
    cqrs_updated = DateTimeField(
        auto_now=True, help_text="This field must be incremented on every model update. "
                                 "It's used to for CQRS sync.",
    )

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        if not self._state.adding:
            self.cqrs_revision = F('cqrs_revision') + 1
        return super(RawMasterMixin, self).save(*args, **kwargs)

    def to_cqrs_dict(self, using=None):
        """ CQRS serialization for transport payload. """
        if self.CQRS_SERIALIZER:
            data = self._class_serialization(using)
        else:
            data = self._common_serialization(using)
        return data

    def cqrs_sync(self, using=None, queue=None):
        """ Manual instance synchronization. """
        if self._state.adding:
            return False

        if not self.CQRS_SERIALIZER:
            try:
                self.refresh_from_db()
            except self._meta.model.DoesNotExist:
                return False

        MasterSignals.post_save(
            self._meta.model, instance=self, using=using, queue=queue, sync=True,
        )
        return True

    def is_sync_instance(self):
        """
        This method can be overridden to apply syncing only to instances by some rules.
        For example, only objects with special status or after some creation date, etc.

        :return: True if this instance needs to be synced, False otherwise
        :rtype: bool
        """
        return True

    @classmethod
    def relate_cqrs_serialization(cls, queryset):
        """
        :param django.db.models.QuerySet queryset:
        """
        return queryset

    @classmethod
    def call_post_bulk_create(cls, instances, using=None):
        """ Post bulk create signal caller (django doesn't support it by default).

        .. code-block:: python
            # On PostgreSQL
            instances = model.objects.bulk_create(instances)
            model.call_post_bulk_create(instances)
        """
        post_bulk_create.send(cls, instances=instances, using=using)

    @classmethod
    def call_post_update(cls, instances, using=None):
        """ Post bulk update signal caller (django doesn't support it by default).

        .. code-block:: python
            # Used automatically by cqrs.bulk_update()
            qs = model.objects.filter(k1=v1)
            model.cqrs.bulk_update(qs, k2=v2)
        """
        post_update.send(cls, instances=instances, using=using)

    def _common_serialization(self, using):
        if isinstance(self.cqrs_revision, CombinedExpression):
            self.refresh_from_db(fields=('cqrs_revision',), using=using)

        opts = self._meta

        if isinstance(self.CQRS_FIELDS, six.string_types) and self.CQRS_FIELDS == ALL_BASIC_FIELDS:
            included_fields = None
        else:
            included_fields = self.CQRS_FIELDS

        data = {}
        for f in chain(opts.concrete_fields, opts.private_fields):
            if included_fields and (f.name not in included_fields):
                continue

            value = f.value_from_object(self)
            if value is not None and isinstance(f, (DateField, DateTimeField)):
                value = str(value)

            data[f.name] = value

        # We need to include additional fields for synchronisation, f.e. to prevent de-duplication
        data['cqrs_revision'] = self.cqrs_revision
        data['cqrs_updated'] = str(self.cqrs_updated)

        return data

    def _class_serialization(self, using):
        db = using if using is not None else self._state.db
        qs = self.__class__._default_manager.using(db).filter(pk=self.pk)

        instance = self.relate_cqrs_serialization(qs).first()
        if not instance:
            raise RuntimeError("Couldn't serialize CQRS class ({}).".format(self.CQRS_ID))

        data = self._cqrs_serializer_cls(instance).data
        data['cqrs_revision'] = instance.cqrs_revision
        data['cqrs_updated'] = str(instance.cqrs_updated)

        return data

    @property
    def _cqrs_serializer_cls(self):
        """ Serialization class loader. """
        if hasattr(self.__class__, '_cqrs_serializer_class'):
            return self.__class__._cqrs_serializer_class

        try:
            serializer = import_string(self.CQRS_SERIALIZER)
            self.__class__._cqrs_serializer_class = serializer
            return serializer
        except ImportError:
            raise ImportError("Model {}: CQRS_SERIALIZER can't be imported.".format(self.__class__))


class MasterMixin(six.with_metaclass(MasterMeta, RawMasterMixin)):
    """
    Mixin for the master CQRS model, that will send data updates to it's replicas.

    CQRS_ID - Unique CQRS identifier for all microservices.
    CQRS_FIELDS - Fields, that need to by synchronized between microservices.
    CQRS_SERIALIZER - Serializer, that overrides common serialization logic.
                      Can't be set together with non-default CQRS_FIELDS.
                      DRF serializers are only supported now: Serializer(instance).data

    cqrs - Manager, that adds needed CQRS queryset methods.
    """
    class Meta:
        abstract = True


class ReplicaMixin(six.with_metaclass(ReplicaMeta, Model)):
    """
    Mixin for the replica CQRS model, that will receive data updates from master. Models, using
    this mixin should be readonly, but this is not enforced (f.e. for admin).

    CQRS_ID - Unique CQRS identifier for all microservices.
    CQRS_MAPPING - Mapping of master data field name to replica model field name.
    CQRS_CUSTOM_SERIALIZATION - To override default data check.
    """
    CQRS_ID = None
    CQRS_MAPPING = None
    CQRS_CUSTOM_SERIALIZATION = False

    objects = Manager()
    cqrs = ReplicaManager()

    cqrs_revision = IntegerField()
    cqrs_updated = DateTimeField()

    class Meta:
        abstract = True

    @classmethod
    def cqrs_save(cls, master_data, sync=False):
        """ This method saves (creates or updates) model instance from CQRS master instance data.
        This method must not be overridden. Otherwise, sync checks need to be implemented manually.

        :param dict master_data: CQRS master instance data.
        :param bool sync: Sync package flag.
        :return: Model instance.
        :rtype: django.db.models.Model
        """
        return cls.cqrs.save_instance(master_data, sync)

    @classmethod
    def cqrs_create(cls, sync, **mapped_data):
        """ This method creates model instance from CQRS mapped instance data. It must be overridden
        by replicas of master models with custom serialization.

        :param dict mapped_data: CQRS mapped instance data.
        :param bool sync: Sync package flag.
        :return: Model instance.
        :rtype: django.db.models.Model
        """
        return cls._default_manager.create(**mapped_data)

    def cqrs_update(self, sync, **mapped_data):
        """ This method updates model instance from CQRS mapped instance data. It must be overridden
        by replicas of master models with custom serialization.

        :param dict mapped_data: CQRS mapped instance data.
        :param bool sync: Sync package flag.
        :return: Model instance.
        :rtype: django.db.models.Model
        """

        for key, value in mapped_data.items():
            setattr(self, key, value)
        self.save()
        return self

    @classmethod
    def cqrs_delete(cls, master_data):
        """ This method deletes model instance from mapped CQRS master instance data.

        :param dict master_data: CQRS master instance data.
        :return: Flag, if delete operation is successful (even if nothing was deleted).
        :rtype: bool
        """
        return cls.cqrs.delete_instance(master_data)



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import pytest

from django.db.models import CharField, IntegerField
from django.utils.timezone import now

from dj_cqrs.metas import ReplicaMeta
from tests.dj_replica import models
from tests.utils import db_error


class ReplicaMetaTest(ReplicaMeta):
    @classmethod
    def check_cqrs_mapping(cls, model_cls):
        return cls._check_cqrs_mapping(model_cls)


def test_cqrs_fields_non_existing_field(mocker):
    with pytest.raises(AssertionError) as e:

        class Cls(object):
            CQRD_ID = 'ID'
            CQRS_MAPPING = {
                'chr_field': 'char_field',
                'integer_field': 'int_field',
            }

            char_field = CharField(max_length=100, primary_key=True)
            int_field = IntegerField()

            _meta = mocker.MagicMock(concrete_fields=(char_field, int_field), private_fields=())
            _meta.pk.name = 'char_field'

        ReplicaMetaTest.check_cqrs_mapping(Cls)

    assert str(e.value) == 'CQRS_MAPPING field is not correctly set for model Cls.'


def test_cqrs_fields_id_is_not_included(mocker):
    with pytest.raises(AssertionError) as e:

        class Cls(object):
            CQRD_ID = 'ID'
            CQRS_MAPPING = {
                'integer_field': 'int_field',
            }

            char_field = CharField(max_length=100, primary_key=True)
            int_field = IntegerField()

            _meta = mocker.MagicMock(concrete_fields=(char_field, int_field), private_fields=())
            _meta.pk.name = 'char_field'

        ReplicaMetaTest.check_cqrs_mapping(Cls)

    assert str(e.value) == 'PK is not in CQRS_MAPPING for model Cls.'


def test_cqrs_fields_duplicates(mocker):
    with pytest.raises(AssertionError) as e:

        class Cls(object):
            CQRD_ID = 'ID'
            CQRS_MAPPING = {
                'integer_field': 'char_field',
                'char_field': 'char_field',
            }

            char_field = CharField(max_length=100, primary_key=True)
            int_field = IntegerField()

            _meta = mocker.MagicMock(concrete_fields=(char_field, int_field), private_fields=())
            _meta.pk.name = 'char_field'

        ReplicaMetaTest.check_cqrs_mapping(Cls)

    assert str(e.value) == 'Duplicate names in CQRS_MAPPING field for model Cls.'


@pytest.mark.django_db
def test_create_simple():
    instance = models.BasicFieldsModelRef.cqrs_save({
        'int_field': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
        'char_field': 'text',
        'bool_field': False,
        'date_field': None,
        'datetime_field': now(),
        'float_field': 1.25,
    })
    assert isinstance(instance, models.BasicFieldsModelRef)

    instance.refresh_from_db()
    assert instance.char_field == 'text'
    assert instance.float_field == 1.25


@pytest.mark.django_db
def test_create_simple_excessive_data():
    instance = models.BasicFieldsModelRef.cqrs_save({
        'int_field': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
        'char_field': 'text',
        'unexpected_field': 'value',
    })
    assert isinstance(instance, models.BasicFieldsModelRef)


@pytest.mark.django_db
def test_create_simple_insufficient_data(caplog):
    models.BasicFieldsModelRef.cqrs_save({
        'int_field': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
    })

    assert 'Not all required CQRS fields are provided in data (basic).' in caplog.text


@pytest.mark.django_db
def test_create_mapped(caplog):
    instance = models.MappedFieldsModelRef.cqrs_save({
        'int_field': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
        'char_field': 'text',
    })
    assert isinstance(instance, models.MappedFieldsModelRef)

    instance.refresh_from_db()
    assert instance.id == 1
    assert instance.name == 'text'


@pytest.mark.django_db
def test_create_mapped_bad_mapping(caplog):
    models.BadMappingModelRef.cqrs_save({
        'int_field': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
        'char_field': 'text',
    })

    assert 'Bad master-replica mapping for invalid_field (basic_3).' in caplog.text


@pytest.mark.django_db
def test_create_db_error(mocker, caplog):
    mocker.patch.object(models.BasicFieldsModelRef.objects, 'create', side_effect=db_error)

    models.BasicFieldsModelRef.cqrs_save({
        'int_field': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
        'char_field': 'text',
    })
    assert 'CQRS create error: pk = 1 (basic).' in caplog.text


@pytest.mark.django_db
def test_update_ok():
    models.BasicFieldsModelRef.objects.create(**{
        'int_field': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
        'char_field': 'text',
    })

    instance = models.BasicFieldsModelRef.cqrs_save({
        'int_field': 1,
        'cqrs_revision': 1,
        'cqrs_updated': now(),
        'char_field': 'new_text',
        'float_field': 1.30,
    })

    assert isinstance(instance, models.BasicFieldsModelRef)

    instance.refresh_from_db()
    assert instance.int_field == 1
    assert instance.char_field == 'new_text'
    assert instance.float_field == 1.30


@pytest.mark.django_db
def test_update_db_error(mocker, caplog):
    models.BasicFieldsModelRef.objects.create(**{
        'int_field': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
        'char_field': 'text',
    })

    mocker.patch.object(models.BasicFieldsModelRef, 'save', side_effect=db_error)

    models.BasicFieldsModelRef.cqrs_save({
        'int_field': 1,
        'cqrs_revision': 1,
        'cqrs_updated': now(),
        'char_field': 'text',
    })
    assert 'CQRS update error: pk = 1, cqrs_revision = 1 (basic).' in caplog.text


@pytest.mark.django_db
def test_delete_ok():
    dt = now()
    models.BasicFieldsModelRef.objects.create(**{
        'int_field': 1,
        'cqrs_revision': 0,
        'cqrs_updated': dt,
        'char_field': 'text',
    })

    is_deleted = models.BasicFieldsModelRef.cqrs_delete({
        'id': 1,
        'cqrs_revision': 0,
        'cqrs_updated': dt,
    })

    assert is_deleted
    assert models.BasicFieldsModelRef.objects.count() == 0


@pytest.mark.django_db
def test_delete_non_existing_id():
    is_deleted = models.BasicFieldsModelRef.cqrs_delete({
        'id': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
    })

    assert is_deleted
    assert models.BasicFieldsModelRef.objects.count() == 0


@pytest.mark.django_db
def test_delete_db_error(mocker, caplog):
    mocker.patch.object(models.BasicFieldsModelRef.objects, 'filter', side_effect=db_error)

    is_deleted = models.BasicFieldsModelRef.cqrs_delete({
        'id': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
    })

    assert not is_deleted
    assert 'CQRS delete error: pk = 1' in caplog.text


@pytest.mark.django_db
def test_save_bad_master_data_field_type(caplog):
    models.BadTypeModelRef.cqrs_save({
        'int_field': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
        'datetime_field': now(),
    })
    assert 'CQRS create error: pk = 1 (basic_1).' in caplog.text


@pytest.mark.django_db
def test_save_no_pk_in_master_data(caplog):
    models.BasicFieldsModelRef.cqrs_save({
        'id': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
        'char_field': 'text',
    })

    assert 'CQRS PK is not provided in data (basic).' in caplog.text


@pytest.mark.django_db
def test_save_no_cqrs_fields_in_master_data(caplog):
    models.BasicFieldsModelRef.cqrs_save({
        'int_field': 1,
        'cqrs_revision': 0,
        'char_field': 'text',
    })

    assert 'CQRS sync fields are not provided in data (basic).' in caplog.text


@pytest.mark.django_db
def test_delete_no_id_in_master_data(caplog):
    is_deleted = models.BasicFieldsModelRef.cqrs_delete({
        'cqrs_revision': 0,
        'cqrs_updated': now(),
    })

    assert not is_deleted
    assert 'CQRS PK is not provided in data (basic).' in caplog.text


@pytest.mark.django_db
def test_delete_no_cqrs_fields_in_master_data(caplog):
    is_deleted = models.BasicFieldsModelRef.cqrs_delete({
        'id': 1,
        'cqrs_revision': 0,
    })

    assert not is_deleted
    assert 'CQRS sync fields are not provided in data (basic).' in caplog.text


@pytest.mark.django_db(transaction=True)
def test_update_before_create_is_over(caplog):
    create_data = {
        'int_field': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
        'char_field': 'text',
    }

    update_data = {
        'int_field': 1,
        'cqrs_revision': 1,
        'cqrs_updated': now(),
        'char_field': 'new_text',
    }

    updated_instance = models.BasicFieldsModelRef.cqrs_save(update_data)
    created_instance = models.BasicFieldsModelRef.cqrs.create_instance(create_data)
    updated_instance.refresh_from_db()

    assert updated_instance.cqrs_revision == 1
    assert updated_instance.char_field == 'new_text'
    assert not created_instance
    assert 'Potentially wrong CQRS sync order: pk = 1, cqrs_revision = 0 (basic).' in caplog.text
    assert 'CQRS create error: pk = 1 (basic).' in caplog.text


@pytest.mark.django_db(transaction=True)
def test_wrong_update_order(caplog):
    models.BasicFieldsModelRef.objects.create(**{
        'int_field': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
        'char_field': 'text',
    })

    update_data_1 = {
        'int_field': 1,
        'cqrs_revision': 1,
        'cqrs_updated': now(),
        'char_field': 'new_text_1',
    }

    update_data_2 = {
        'int_field': 1,
        'cqrs_revision': 2,
        'cqrs_updated': now(),
        'char_field': 'new_text_2',
    }

    earlier_instance = models.BasicFieldsModelRef.cqrs_save(update_data_2)
    later_instance = models.BasicFieldsModelRef.cqrs_save(update_data_1)
    earlier_instance.refresh_from_db()

    assert earlier_instance.cqrs_revision == 2
    assert earlier_instance.char_field == 'new_text_2'
    assert later_instance
    assert 'Wrong CQRS sync order: pk = 1, cqrs_revision = new 1 / existing 2 (basic).' in \
        caplog.text


@pytest.mark.django_db(transaction=True)
def test_de_duplication(caplog):
    models.BasicFieldsModelRef.objects.create(**{
        'int_field': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
        'char_field': 'text',
    })

    update_data = {
        'int_field': 1,
        'cqrs_revision': 1,
        'cqrs_updated': now(),
        'char_field': 'new_text',
    }

    earlier_instance = models.BasicFieldsModelRef.cqrs_save(update_data)
    duplicate_instance = models.BasicFieldsModelRef.cqrs_save(update_data)

    assert earlier_instance.cqrs_revision == 1
    assert earlier_instance.char_field == 'new_text'
    assert duplicate_instance.cqrs_revision == 1
    assert duplicate_instance.char_field == 'new_text'
    assert 'Received duplicate CQRS data: pk = 1, cqrs_revision = 1 (basic).' in caplog.text


@pytest.mark.django_db(transaction=True)
def test_create_before_delete_is_over(caplog):
    # This situation may extremely rarely happen, if the IDs are not auto incremented on master
    #  and are not unique in the infinite timeline.
    # This will lead to expected inconsistency.

    models.BasicFieldsModelRef.objects.create(**{
        'int_field': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
        'char_field': 'text',
    })

    delete_data = {
        'id': 1,
        'cqrs_revision': 1,
        'cqrs_updated': now(),
    }
    new_create_data = {
        'int_field': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
        'char_field': 'other',
    }

    models.BasicFieldsModelRef.cqrs_save(new_create_data)
    is_deleted = models.BasicFieldsModelRef.cqrs_delete(delete_data)

    assert 'Received duplicate CQRS data: pk = 1, cqrs_revision = 0 (basic).' in caplog.text
    assert 'CQRS potential creation race condition: pk = 1 (basic).' in caplog.text
    assert is_deleted


@pytest.mark.django_db
def test_updates_were_lost(caplog):
    models.BasicFieldsModelRef.objects.create(**{
        'int_field': 1,
        'cqrs_revision': 0,
        'cqrs_updated': now(),
        'char_field': 'text',
    })

    models.BasicFieldsModelRef.cqrs_save({
        'int_field': 1,
        'cqrs_revision': 5,
        'cqrs_updated': now(),
        'char_field': 'text1',
    })

    assert 'Lost or filtered out 4 CQRS packages: pk = 1, cqrs_revision = 5 (basic)' in caplog.text



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import pytest

from integration_tests.tests.utils import (
    REPLICA_BASIC_TABLE, count_replica_rows, get_replica_all, get_replica_first, transport_delay,
)
from tests.dj_master.models import BasicFieldsModel


@pytest.mark.django_db(transaction=True)
def test_flow(replica_cursor):
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 0

    # Create
    master_instances = BasicFieldsModel.objects.bulk_create([
        BasicFieldsModel(
            int_field=index,
            char_field='text',
        )
        for index in range(1, 4)
    ])
    BasicFieldsModel.call_post_bulk_create(master_instances)

    transport_delay()
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 3

    assert {'text'} == {
        t[0] for t in get_replica_all(replica_cursor, REPLICA_BASIC_TABLE, ('char_field',))
    }

    # Update 1 and 2
    BasicFieldsModel.cqrs.bulk_update(
        BasicFieldsModel.objects.filter(int_field__in=(1, 2)),
        char_field='new_text',
    )

    transport_delay()
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 3

    assert ['new_text', 'new_text', 'text'] == [
        t[0] for t in get_replica_all(
            replica_cursor, REPLICA_BASIC_TABLE, ('char_field',), order_asc_by='int_field',
        )
    ]

    # Delete 1 and 3
    BasicFieldsModel.objects.filter(int_field__in=(1, 3)).delete()

    transport_delay()
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 1

    assert (2, 'new_text', 1) == get_replica_first(
        replica_cursor, REPLICA_BASIC_TABLE, ('int_field', 'char_field', 'cqrs_revision'),
    )



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import logging
import time
from socket import gaierror

import ujson
from django.conf import settings
from pika import exceptions, BasicProperties, BlockingConnection, ConnectionParameters, credentials

from dj_cqrs.constants import SignalType
from dj_cqrs.controller import consumer
from dj_cqrs.dataclasses import TransportPayload
from dj_cqrs.registries import ReplicaRegistry
from dj_cqrs.transport import BaseTransport

logger = logging.getLogger()


class RabbitMQTransport(BaseTransport):
    CONSUMER_RETRY_TIMEOUT = 5

    @classmethod
    def consume(cls):
        consumer_rabbit_settings = cls._get_consumer_settings()
        common_rabbit_settings = cls._get_common_settings()

        while True:
            connection = None
            try:
                connection, channel = cls._get_consumer_rmq_objects(
                    *(common_rabbit_settings + consumer_rabbit_settings)
                )
                channel.start_consuming()
            except (exceptions.AMQPError,
                    exceptions.ChannelError,
                    exceptions.ReentrancyError,
                    gaierror):
                logger.error('AMQP connection error. Reconnecting...')
                time.sleep(cls.CONSUMER_RETRY_TIMEOUT)
            finally:
                if connection:
                    connection.close()

    @classmethod
    def produce(cls, payload):
        rmq_settings = cls._get_common_settings()
        exchange = rmq_settings[-1]

        connection = None
        try:
            # Decided not to create context-manager to stay within the class
            connection, channel = cls._get_producer_rmq_objects(*rmq_settings)

            cls._produce_message(channel, exchange, payload)
            cls._log_produced(payload)
        except (exceptions.AMQPError, exceptions.ChannelError, exceptions.ReentrancyError):
            logger.error("CQRS couldn't be published: pk = {} ({}).".format(
                payload.pk, payload.cqrs_id,
            ))
        finally:
            if connection:
                connection.close()

    @classmethod
    def _consume_message(cls, ch, method, properties, body):
        try:
            dct = ujson.loads(body)
            for key in ('signal_type', 'cqrs_id', 'instance_data'):
                if key not in dct:
                    raise ValueError

            if 'instance_pk' not in dct:
                logger.warning('CQRS deprecated package structure.')

        except ValueError:
            logger.error("CQRS couldn't be parsed: {}.".format(body))
            ch.basic_reject(delivery_tag=method.delivery_tag, requeue=False)
            return

        payload = TransportPayload(
            dct['signal_type'], dct['cqrs_id'], dct['instance_data'], dct.get('instance_pk'),
        )

        cls._log_consumed(payload)
        instance = consumer.consume(payload)

        if instance:
            ch.basic_ack(delivery_tag=method.delivery_tag)
            cls._log_consumed_accepted(payload)
        else:
            ch.basic_nack(delivery_tag=method.delivery_tag, requeue=False)
            cls._log_consumed_denied(payload)

    @classmethod
    def _produce_message(cls, channel, exchange, payload):
        routing_key = cls._get_produced_message_routing_key(payload)

        channel.basic_publish(
            exchange=exchange,
            routing_key=routing_key,
            body=ujson.dumps(payload.to_dict()),
            mandatory=True,
            properties=BasicProperties(content_type='text/plain', delivery_mode=2)
        )

    @staticmethod
    def _get_produced_message_routing_key(payload):
        routing_key = payload.cqrs_id

        if payload.signal_type == SignalType.SYNC and payload.queue:
            routing_key = 'cqrs.{}.{}'.format(payload.queue, routing_key)

        return routing_key

    @classmethod
    def _get_consumer_rmq_objects(cls, host, port, creds, exchange, queue_name, prefetch_count):
        connection = BlockingConnection(
            ConnectionParameters(host=host, port=port, credentials=creds),
        )
        channel = connection.channel()
        channel.basic_qos(prefetch_count=prefetch_count)

        cls._declare_exchange(channel, exchange)
        channel.queue_declare(queue_name, durable=True, exclusive=False)

        for cqrs_id, replica_model in ReplicaRegistry.models.items():
            channel.queue_bind(exchange=exchange, queue=queue_name, routing_key=cqrs_id)

            # Every service must have specific SYNC routes
            channel.queue_bind(
                exchange=exchange,
                queue=queue_name,
                routing_key='cqrs.{}.{}'.format(queue_name, cqrs_id),
            )

        channel.basic_consume(
            queue=queue_name,
            on_message_callback=cls._consume_message,
            auto_ack=False,
            exclusive=False,
        )

        return connection, channel

    @classmethod
    def _get_producer_rmq_objects(cls, host, port, creds, exchange):
        connection = BlockingConnection(
            ConnectionParameters(
                host=host,
                port=port,
                credentials=creds,
                blocked_connection_timeout=10,
            ),
        )
        channel = connection.channel()
        cls._declare_exchange(channel, exchange)

        return connection, channel

    @staticmethod
    def _declare_exchange(channel, exchange):
        channel.exchange_declare(
            exchange=exchange,
            exchange_type='topic',
            durable=True,
        )

    @staticmethod
    def _get_common_settings():
        host = settings.CQRS.get('host', ConnectionParameters.DEFAULT_HOST)
        port = settings.CQRS.get('port', ConnectionParameters.DEFAULT_PORT)
        user = settings.CQRS.get('user', ConnectionParameters.DEFAULT_USERNAME)
        password = settings.CQRS.get('password', ConnectionParameters.DEFAULT_PASSWORD)
        exchange = settings.CQRS.get('exchange', 'cqrs')
        return (
            host,
            port,
            credentials.PlainCredentials(user, password, erase_on_connect=True),
            exchange,
        )

    @staticmethod
    def _get_consumer_settings():
        queue_name = settings.CQRS['queue']
        consumer_prefetch_count = settings.CQRS.get('consumer_prefetch_count', 10)
        return (
            queue_name,
            consumer_prefetch_count,
        )

    @staticmethod
    def _log_consumed(payload):
        """
        :param dj_cqrs.dataclasses.TransportPayload payload: Transport payload from master model.
        """
        if payload.pk:
            logger.info('CQRS is received: pk = {} ({}).'.format(payload.pk, payload.cqrs_id))

    @staticmethod
    def _log_consumed_accepted(payload):
        """
        :param dj_cqrs.dataclasses.TransportPayload payload: Transport payload from master model.
        """
        if payload.pk:
            logger.info('CQRS is applied: pk = {} ({}).'.format(payload.pk, payload.cqrs_id))

    @staticmethod
    def _log_consumed_denied(payload):
        """
        :param dj_cqrs.dataclasses.TransportPayload payload: Transport payload from master model.
        """
        if payload.pk:
            logger.info('CQRS is denied: pk = {} ({}).'.format(payload.pk, payload.cqrs_id))

    @staticmethod
    def _log_produced(payload):
        """
        :param dj_cqrs.dataclasses.TransportPayload payload: Transport payload from master model.
        """
        logger.info('CQRS is published: pk = {} ({}).'.format(payload.pk, payload.cqrs_id))



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import pytest
from django.core.management import CommandError, call_command
from tests.utils import db_error

from dj_cqrs.constants import SignalType
from tests.dj_master.models import Author

COMMAND_NAME = 'cqrs_sync'


def test_no_cqrs_id():
    with pytest.raises(CommandError):
        call_command(COMMAND_NAME)


def test_bad_cqrs_id():
    with pytest.raises(CommandError) as e:
        call_command(COMMAND_NAME, '--cqrs-id=invalid')

    assert 'Wrong CQRS ID: invalid!' in str(e)


def test_empty_filter_arg(capsys):
    call_command(COMMAND_NAME, '--cqrs-id=author')

    captured = capsys.readouterr()
    assert 'No objects found for filter!' in captured.out


def test_unparseable_filter():
    with pytest.raises(CommandError) as e:
        call_command(COMMAND_NAME, '--cqrs-id=author', '-f={arg}')

    assert 'Bad filter kwargs!' in str(e)


def test_non_kwargs_filter():
    with pytest.raises(CommandError) as e:
        call_command(COMMAND_NAME, '--cqrs-id=author', '-f=[1]')

    assert 'Bad filter kwargs!' in str(e)


def test_bad_kwargs_filter():
    with pytest.raises(CommandError) as e:
        call_command(COMMAND_NAME, '--cqrs-id=author', '-f={"field": "value"}')

    assert 'Bad filter kwargs! Cannot resolve keyword' in str(e)


@pytest.mark.django_db
def test_empty_filter(capsys):
    call_command(COMMAND_NAME, '--cqrs-id=author', '-f={"id": 1}')

    captured = capsys.readouterr()
    assert 'No objects found for filter!' in captured.out


@pytest.mark.django_db(transaction=True)
def test_no_queue(mocker, capsys):
    Author.objects.create(id=1, name='author')

    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')
    call_command(COMMAND_NAME, '--cqrs-id=author', '-f={"id": 1}')

    publisher_mock.assert_called_once()
    payload = publisher_mock.call_args[0][0]
    assert payload.queue is None
    assert payload.signal_type is SignalType.SYNC

    captured = capsys.readouterr()
    assert 'Done!\n1 instance(s) synced.\n1 instance(s) processed.' in captured.out


@pytest.mark.django_db(transaction=True)
def test_queue_is_set(mocker, capsys):
    Author.objects.create(id=1, name='author')

    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')
    call_command(COMMAND_NAME, '--cqrs-id=author', '-f={"id": 1}', '--queue=replica')

    publisher_mock.assert_called_once()
    payload = publisher_mock.call_args[0][0]
    assert payload.queue == 'replica'
    assert payload.signal_type is SignalType.SYNC

    captured = capsys.readouterr()
    assert 'Done!\n1 instance(s) synced.\n1 instance(s) processed.' in captured.out


@pytest.mark.django_db(transaction=True)
def test_several_synced(mocker, capsys):
    for i in range(1, 3):
        Author.objects.create(id=i, name='author')

    publisher_mock = mocker.patch('dj_cqrs.controller.producer.produce')
    call_command(COMMAND_NAME, '--cqrs-id=author', '-f={"id__in": [1, 2]}')

    assert publisher_mock.call_count == 2

    captured = capsys.readouterr()
    assert 'Done!\n2 instance(s) synced.\n2 instance(s) processed.' in captured.out


@pytest.mark.django_db
def test_error(capsys, mocker):
    Author.objects.create(id=2, name='2')

    mocker.patch('tests.dj_master.models.Author.cqrs_sync', side_effect=db_error)
    call_command(COMMAND_NAME, '--cqrs-id=author', '-f={}')

    captured = capsys.readouterr()
    assert 'Sync record failed for pk=2' in captured.out
    assert '1 instance(s) processed.' in captured.out
    assert '0 instance(s) synced.' in captured.out


@pytest.mark.django_db
def test_progress(capsys):
    Author.objects.create(id=2, name='2')
    call_command(COMMAND_NAME, '--cqrs-id=author', '--progress', '-f={}', '--batch=2')

    captured = capsys.readouterr()
    assert 'Processing 1 records with batch size 2' in captured.out
    assert '1 of 1 processed - 100% with rate' in captured.out



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import logging

from django.conf import settings

logger = logging.getLogger()


class RegistryMixin:
    @classmethod
    def register_model(cls, model_cls):
        """ Registration of CQRS model identifiers. """
        assert model_cls.CQRS_ID not in cls.models, "Two models can't have the same CQRS_ID: {}." \
            .format(model_cls.CQRS_ID)
        cls.models[model_cls.CQRS_ID] = model_cls

    @classmethod
    def get_model_by_cqrs_id(cls, cqrs_id):
        if cqrs_id in cls.models:
            return cls.models[cqrs_id]

        logger.error('No model with such CQRS_ID: {}.'.format(cqrs_id))


class MasterRegistry(RegistryMixin):
    models = {}


class ReplicaRegistry(RegistryMixin):
    models = {}

    @classmethod
    def register_model(cls, model_cls):
        assert getattr(settings, 'CQRS', {}).get('queue') is not None, \
            'CQRS queue must be set for the service, that has replica models.'
        super(ReplicaRegistry, cls).register_model(model_cls)



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import logging
from itertools import chain

from django.core.exceptions import ValidationError
from django.db import Error, IntegrityError, transaction
from django.db.models import Manager, F
from django.utils import timezone


logger = logging.getLogger()


class MasterManager(Manager):
    def bulk_update(self, queryset, **kwargs):
        """ Custom update method to support sending update signals.

        :param django.db.models.QuerySet queryset: Django Queryset (f.e. filter)
        :param kwargs: Update kwargs
        """
        with transaction.atomic():
            current_dt = timezone.now()
            result = queryset.update(
                cqrs_revision=F('cqrs_revision') + 1, cqrs_updated=current_dt, **kwargs
            )
        queryset.model.call_post_update(list(queryset.all()), using=queryset.db)
        return result


class ReplicaManager(Manager):
    def save_instance(self, master_data, sync=False):
        """ This method saves (creates or updates) model instance from CQRS master instance data.

        :param dict master_data: CQRS master instance data.
        :param bool sync: Sync package flag.
        :return: Model instance.
        :rtype: django.db.models.Model
        """
        mapped_data = self._map_save_data(master_data)

        if mapped_data:
            pk_name = self._get_model_pk_name()
            pk_value = mapped_data[pk_name]
            f_kwargs = {pk_name: pk_value}

            instance = self.model._default_manager.filter(**f_kwargs).first()

            if instance:
                return self.update_instance(instance, mapped_data, sync=sync)

            return self.create_instance(mapped_data, sync=sync)

    def create_instance(self, mapped_data, sync=False):
        """ This method creates model instance from mapped CQRS master instance data.

        :param dict mapped_data: Mapped CQRS master instance data.
        :param bool sync: Sync package flag.
        :return: ReplicaMixin model instance.
        :rtype: django.db.models.Model
        """
        try:
            return self.model.cqrs_create(sync, **mapped_data)
        except (Error, ValidationError) as e:
            pk_value = mapped_data[self._get_model_pk_name()]
            if isinstance(e, IntegrityError):
                logger.warning(
                    'Potentially wrong CQRS sync order: pk = {}, cqrs_revision = {} ({}).'.format(
                        pk_value, mapped_data['cqrs_revision'], self.model.CQRS_ID,
                    ),
                )

            logger.error(
                '{}\nCQRS create error: pk = {} ({}).'.format(
                    str(e), pk_value, self.model.CQRS_ID,
                ),
            )

    def update_instance(self, instance, mapped_data, sync=False):
        """ This method updates model instance from mapped CQRS master instance data.

        :param django.db.models.Model instance: ReplicaMixin model instance.
        :param dict mapped_data: Mapped CQRS master instance data.
        :param bool sync: Sync package flag.
        :return: ReplicaMixin model instance.
        :rtype: django.db.models.Model
        """
        pk_value = mapped_data[self._get_model_pk_name()]
        current_cqrs_revision = mapped_data['cqrs_revision']
        existing_cqrs_revision = instance.cqrs_revision

        if sync:
            if existing_cqrs_revision > current_cqrs_revision:
                w_tpl = 'CQRS revision downgrade on sync: pk = {}, ' \
                    'cqrs_revision = new {} / existing {} ({}).'
                logger.warning(w_tpl.format(
                    pk_value, current_cqrs_revision, existing_cqrs_revision, self.model.CQRS_ID,
                ))

        else:
            if existing_cqrs_revision > current_cqrs_revision:
                e_tpl = 'Wrong CQRS sync order: pk = {}, cqrs_revision = new {} / existing {} ({}).'
                logger.error(e_tpl.format(
                    pk_value, current_cqrs_revision, existing_cqrs_revision, self.model.CQRS_ID,
                ))
                return instance

            if existing_cqrs_revision == current_cqrs_revision:
                logger.error(
                    'Received duplicate CQRS data: pk = {}, cqrs_revision = {} ({}).'.format(
                        pk_value, current_cqrs_revision, self.model.CQRS_ID,
                    ),
                )
                if current_cqrs_revision == 0:
                    logger.warning(
                        'CQRS potential creation race condition: pk = {} ({}).'.format(
                            pk_value, self.model.CQRS_ID,
                        ),
                    )

                return instance

            if current_cqrs_revision != instance.cqrs_revision + 1:
                w_tpl = 'Lost or filtered out {} CQRS packages: pk = {}, cqrs_revision = {} ({})'
                logger.warning(w_tpl.format(
                    current_cqrs_revision - instance.cqrs_revision - 1,
                    pk_value, current_cqrs_revision, self.model.CQRS_ID,
                ))

        try:
            return instance.cqrs_update(sync, **mapped_data)
        except (Error, ValidationError) as e:
            logger.error(
                '{}\nCQRS update error: pk = {}, cqrs_revision = {} ({}).'.format(
                    str(e), pk_value, current_cqrs_revision, self.model.CQRS_ID,
                ),
            )

    def delete_instance(self, master_data):
        """ This method deletes model instance from mapped CQRS master instance data.

        :param dict master_data: CQRS master instance data.
        :return: Flag, if delete operation is successful (even if nothing was deleted).
        :rtype: bool
        """
        mapped_data = self._map_delete_data(master_data)

        if mapped_data:
            pk_name = self._get_model_pk_name()
            pk_value = mapped_data[pk_name]
            try:
                self.model._default_manager.filter(**{pk_name: pk_value}).delete()
                return True
            except Error as e:
                logger.error(
                    '{}\nCQRS delete error: pk = {} ({}).'.format(
                        str(e), pk_value, self.model.CQRS_ID,
                    ),
                )

        return False

    def _map_save_data(self, master_data):
        if not self._cqrs_fields_are_filled(master_data):
            return

        mapped_data = self._make_initial_mapping(master_data)
        if not mapped_data:
            return

        if self._get_model_pk_name() not in mapped_data:
            self._log_pk_data_error()
            return

        if self.model.CQRS_CUSTOM_SERIALIZATION:
            return mapped_data

        mapped_data = self._remove_excessive_data(mapped_data)

        if self._all_required_fields_are_filled(mapped_data):
            return mapped_data

    def _make_initial_mapping(self, master_data):
        if self.model.CQRS_MAPPING is None:
            return master_data

        mapped_data = {
            'cqrs_revision': master_data['cqrs_revision'],
            'cqrs_updated': master_data['cqrs_updated'],
        }
        for master_name, replica_name in self.model.CQRS_MAPPING.items():
            if master_name not in master_data:
                logger.error('Bad master-replica mapping for {} ({}).'.format(
                    master_name, self.model.CQRS_ID,
                ))
                return

            mapped_data[replica_name] = master_data[master_name]
        return mapped_data

    def _remove_excessive_data(self, data):
        opts = self.model._meta
        possible_field_names = {f.name for f in chain(opts.concrete_fields, opts.private_fields)}
        return {k: v for k, v in data.items() if k in possible_field_names}

    def _all_required_fields_are_filled(self, mapped_data):
        opts = self.model._meta

        required_field_names = {
            f.name for f in chain(opts.concrete_fields, opts.private_fields) if not f.null
        }
        if not (required_field_names - set(mapped_data.keys())):
            return True

        logger.error(
            'Not all required CQRS fields are provided in data ({}).'.format(self.model.CQRS_ID),
        )
        return False

    def _map_delete_data(self, master_data):
        if 'id' not in master_data:
            self._log_pk_data_error()
            return

        if not self._cqrs_fields_are_filled(master_data):
            return

        return {
            self._get_model_pk_name(): master_data['id'],
            'cqrs_revision': master_data['cqrs_revision'],
            'cqrs_updated': master_data['cqrs_updated'],
        }

    def _cqrs_fields_are_filled(self, data):
        if 'cqrs_revision' in data and 'cqrs_updated' in data:
            return True

        logger.error('CQRS sync fields are not provided in data ({}).'.format(self.model.CQRS_ID))
        return False

    def _log_pk_data_error(self):
        logger.error('CQRS PK is not provided in data ({}).'.format(self.model.CQRS_ID))

    def _get_model_pk_name(self):
        return self.model._meta.pk.name



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

from rest_framework.serializers import CharField, ModelSerializer

from tests.dj_master.models import Author, Book, Publisher


class BookSerializer(ModelSerializer):
    name = CharField(source='title')

    class Meta:
        model = Book
        fields = ('id', 'name')


class PublisherSerializer(ModelSerializer):
    class Meta:
        model = Publisher
        fields = '__all__'


class AuthorSerializer(ModelSerializer):
    books = BookSerializer(many=True)
    publisher = PublisherSerializer()

    class Meta:
        model = Author
        fields = ('id', 'name', 'publisher', 'books')



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import time


REPLICA_BASIC_TABLE = 'dj_replica_basicfieldsmodelref'
REPLICA_EVENT_TABLE = 'dj_replica_event'

REPLICA_TABLES = (REPLICA_BASIC_TABLE, REPLICA_EVENT_TABLE)


def count_replica_rows(cursor, table):
    cursor.execute('SELECT COUNT(*) FROM {};'.format(table))
    return cursor.fetchone()[0]


def get_replica_all(cursor, table, columns=None, order_asc_by=None):
    select = ','.join(columns) if columns else '*'

    sql = 'SELECT {} FROM {}'.format(select, table)
    if order_asc_by:
        sql = '{} ORDER BY {} ASC;'.format(sql, order_asc_by)

    cursor.execute(sql)
    return cursor.fetchall()


def get_replica_first(cursor, table, columns=None):
    select = ','.join(columns) if columns else '*'

    cursor.execute('SELECT {} FROM {} LIMIT 1;'.format(select, table))
    return cursor.fetchone()


def transport_delay(delay=1):
    time.sleep(delay)



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import pytest

from integration_tests.tests.utils import (
    REPLICA_BASIC_TABLE, count_replica_rows, get_replica_first, transport_delay,
)
from tests.dj_master.models import BasicFieldsModel


@pytest.mark.django_db(transaction=True)
def test_flow(replica_cursor, mocker):
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 0

    # Create
    master_instance = BasicFieldsModel.objects.create(
        int_field=1,
        char_field='text',
    )
    assert master_instance.cqrs_revision == 0

    transport_delay()
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 1

    replica_tuple = get_replica_first(
        replica_cursor, REPLICA_BASIC_TABLE,
        ('int_field', 'char_field', 'cqrs_revision', 'cqrs_updated'),
    )
    assert (
        master_instance.int_field,
        master_instance.char_field,
        master_instance.cqrs_revision,
        master_instance.cqrs_updated,
    ) == replica_tuple

    # We simulate transport error
    mocker.patch('dj_cqrs.controller.producer.produce')
    master_instance.char_field = 'new'
    master_instance.save()
    mocker.stopall()

    master_instance.refresh_from_db()
    assert master_instance.cqrs_revision == 1

    # Sync to other service
    master_instance.cqrs_sync(queue='other_replica')

    transport_delay()
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 1

    replica_tuple = get_replica_first(
        replica_cursor, REPLICA_BASIC_TABLE,
        ('int_field', 'char_field', 'cqrs_revision', 'cqrs_updated'),
    )
    assert replica_tuple[0] == 1
    assert replica_tuple[1] == 'text'
    assert replica_tuple[2] == 0

    # Sync to replica
    master_instance.cqrs_sync(queue='replica')

    transport_delay()
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 1

    replica_tuple = get_replica_first(
        replica_cursor, REPLICA_BASIC_TABLE,
        ('int_field', 'char_field', 'cqrs_revision', 'cqrs_updated'),
    )
    assert replica_tuple[0] == 1
    assert replica_tuple[1] == 'new'
    assert replica_tuple[2] == 1

    mocker.patch('dj_cqrs.controller.producer.produce')
    master_instance.char_field = 'new2'
    master_instance.save()
    mocker.stopall()

    # Sync to all
    master_instance.cqrs_sync()

    transport_delay()
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 1

    replica_tuple = get_replica_first(
        replica_cursor, REPLICA_BASIC_TABLE,
        ('int_field', 'char_field', 'cqrs_revision', 'cqrs_updated'),
    )
    assert replica_tuple[0] == 1
    assert replica_tuple[1] == 'new2'
    assert replica_tuple[2] == 2



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import pytest

from dj_cqrs.metas import _MetaUtils


@pytest.mark.django_db
def test_no_cqrs_id():
    with pytest.raises(AssertionError) as e:

        class Cls(object):
            CQRS_ID = None

        _MetaUtils.check_cqrs_id(Cls)

    assert str(e.value) == 'CQRS_ID must be set for every model, that uses CQRS.'



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import pytest
from django.utils.timezone import now

from integration_tests.tests.utils import (
    REPLICA_BASIC_TABLE, count_replica_rows, get_replica_first, transport_delay,
)
from tests.dj_master.models import BasicFieldsModel


@pytest.mark.django_db(transaction=True)
def test_flow(replica_cursor):
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 0

    # Create
    master_instance = BasicFieldsModel.objects.create(
        int_field=1,
        char_field='text',
        date_field=now().date(),
    )
    assert master_instance.cqrs_revision == 0

    transport_delay()
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 1

    replica_tuple = get_replica_first(
        replica_cursor, REPLICA_BASIC_TABLE,
        ('int_field', 'char_field', 'date_field', 'cqrs_revision', 'cqrs_updated', 'bool_field'),
    )
    assert (
        master_instance.int_field,
        master_instance.char_field,
        master_instance.date_field,
        master_instance.cqrs_revision,
        master_instance.cqrs_updated,
        master_instance.bool_field,
    ) == replica_tuple

    # Update
    master_instance.bool_field = True
    master_instance.save()
    master_instance.refresh_from_db()
    assert master_instance.cqrs_revision == 1

    transport_delay()
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 1

    updated_replica_tuple = get_replica_first(
        replica_cursor, REPLICA_BASIC_TABLE,
        ('int_field', 'cqrs_revision', 'cqrs_updated', 'bool_field'),
    )
    assert (
       master_instance.int_field,
       master_instance.cqrs_revision,
       master_instance.cqrs_updated,
       master_instance.bool_field,
    ) == updated_replica_tuple

    # Delete
    master_instance.delete()
    transport_delay()
    assert count_replica_rows(replica_cursor, REPLICA_BASIC_TABLE) == 0



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import logging

from django.db import transaction

from dj_cqrs.constants import SignalType
from dj_cqrs.registries import ReplicaRegistry


logger = logging.getLogger()


def consume(payload):
    """ Consumer controller.

    :param dj_cqrs.dataclasses.TransportPayload payload: Consumed payload from master service.
    """
    return route_signal_to_replica_model(
        payload.signal_type, payload.cqrs_id, payload.instance_data,
    )


def route_signal_to_replica_model(signal_type, cqrs_id, instance_data):
    """ Routes signal to model method to create/update/delete replica instance.

    :param dj_cqrs.constants.SignalType signal_type: Consumed signal type.
    :param str cqrs_id: Replica model CQRS unique identifier.
    :param dict instance_data: Master model data.
    """
    model_cls = ReplicaRegistry.get_model_by_cqrs_id(cqrs_id)

    if model_cls:
        if signal_type == SignalType.DELETE:
            with transaction.atomic():
                return model_cls.cqrs_delete(instance_data)

        elif signal_type == SignalType.SAVE:
            with transaction.atomic():
                return model_cls.cqrs_save(instance_data)

        elif signal_type == SignalType.SYNC:
            with transaction.atomic():
                return model_cls.cqrs_save(instance_data, sync=True)

        else:
            logger.error('Bad signal type "{}" for CQRS_ID "{}".'.format(signal_type, cqrs_id))



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

from setuptools import find_packages, setup


def read_file(name):
    with open(name, 'r') as f:
        content = f.read().rstrip('\n')
    return content


setup(
    name='django-cqrs',
    author='Ingram Micro',
    url='https://connect.cloud.im',
    description='Django CQRS data synchronisation',
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    license='Apache License, Version 2.0',

    python_requires='>=3.6',
    zip_safe=True,
    packages=find_packages(),
    include_package_data=True,
    install_requires=read_file('requirements/dev.txt').splitlines(),
    tests_require=read_file('requirements/test.txt').splitlines(),
    setup_requires=['setuptools_scm', 'pytest-runner'],
    use_scm_version=True,

    keywords='django cqrs sql mixin amqp',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Django :: 1.11',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Communications',
        'Topic :: Database',
    ]
)



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.


def batch_qs(qs, batch_size=10000):
    """
    Helper function to manage RAM usage on big dataset iterations.
    This function can be used only on STATIC DB state. It's a good fit for migrations, but
    it can't be used in real applications.
    """
    assert batch_size > 0

    total = qs.count()
    for start in range(0, total, batch_size):
        end = min(start + batch_size, total)
        yield qs[start:end]



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

import ujson

import pytest
from django.core.management import CommandError, call_command

from tests.dj_master.models import Author


COMMAND_NAME = 'cqrs_diff_master'


def test_no_cqrs_id():
    with pytest.raises(CommandError):
        call_command(COMMAND_NAME)


def test_bad_cqrs_id():
    with pytest.raises(CommandError) as e:
        call_command(COMMAND_NAME, '--cqrs-id=invalid')

    assert 'Wrong CQRS ID: invalid!' in str(e)


@pytest.mark.django_db
def test_first_row(capsys):
    Author.objects.create(name='author', id=1)

    call_command(COMMAND_NAME, '--cqrs-id=author')

    captured = capsys.readouterr()
    assert '{},'.format(Author.CQRS_ID) in captured.out


@pytest.mark.django_db
def test_objects_less_than_batch(capsys):
    author = Author.objects.create(name='author', id=1)

    call_command(COMMAND_NAME, '--cqrs-id=author')

    captured = capsys.readouterr()
    out_lines = captured.out.split('\n')

    assert ujson.loads(out_lines[1]) == [[author.pk, author.cqrs_revision]]


@pytest.mark.django_db
def test_objects_more_than_batch(capsys):
    for i in range(3):
        Author.objects.create(name=str(i), id=i)

    call_command(COMMAND_NAME, '--cqrs-id=author', '--batch=2')

    captured = capsys.readouterr()
    out_lines = captured.out.split('\n')
    assert ujson.loads(out_lines[1]) == [
        [author.pk, author.cqrs_revision] for author in Author.objects.all()[:2]
    ]
    assert ujson.loads(out_lines[2]) == [
        [author.pk, author.cqrs_revision] for author in Author.objects.all()[2:]
    ]


@pytest.mark.django_db
def test_filter_no_objects(capsys):
    call_command(COMMAND_NAME, '--cqrs-id=author', '-f={"id__in": [1, 2]}')

    captured = capsys.readouterr()
    assert 'No objects found for filter!' in captured.err


@pytest.mark.django_db
def test_objects_are_filtered(capsys):
    for i in range(2):
        Author.objects.create(name=str(i), id=i)

    call_command(COMMAND_NAME, '--cqrs-id=author', '-f={"id__in": [1, 3]}')

    captured = capsys.readouterr()
    out_lines = captured.out.split('\n')

    assert ujson.loads(out_lines[1]) == [[1, 0]]



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

from itertools import chain

from django.db.models import base

from dj_cqrs.constants import ALL_BASIC_FIELDS
from dj_cqrs.registries import MasterRegistry, ReplicaRegistry
from dj_cqrs.signals import MasterSignals


class MasterMeta(base.ModelBase):
    def __new__(mcs, *args):
        model_cls = super(MasterMeta, mcs).__new__(mcs, *args)

        if args[0] != 'MasterMixin':
            mcs.register(model_cls)

        return model_cls

    @staticmethod
    def register(model_cls):
        _MetaUtils.check_cqrs_id(model_cls)
        MasterMeta._check_correct_configuration(model_cls)
        if model_cls.CQRS_SERIALIZER is None:
            MasterMeta._check_cqrs_fields(model_cls)

        MasterRegistry.register_model(model_cls)
        MasterSignals.register_model(model_cls)
        return model_cls

    @staticmethod
    def _check_correct_configuration(model_cls):
        """ Check that model has correct CQRS configuration.

        :param dj_cqrs.mixins.MasterMixin model_cls: CQRS Master Model.
        :raises: AssertionError
        """
        if model_cls.CQRS_FIELDS != ALL_BASIC_FIELDS:
            assert model_cls.CQRS_SERIALIZER is None, \
                "Model {}: CQRS_FIELDS can't be set together with CQRS_SERIALIZER.".format(
                    model_cls,
                )

    @staticmethod
    def _check_cqrs_fields(model_cls):
        """ Check that model has correct CQRS fields configuration.

        :param dj_cqrs.mixins.MasterMixin model_cls: CQRS Master Model.
        :raises: AssertionError
        """
        if model_cls.CQRS_FIELDS != ALL_BASIC_FIELDS:
            cqrs_field_names = list(model_cls.CQRS_FIELDS)
            _MetaUtils.check_cqrs_field_setting(model_cls, cqrs_field_names, 'CQRS_FIELDS')


class ReplicaMeta(base.ModelBase):
    def __new__(mcs, *args):
        model_cls = super(ReplicaMeta, mcs).__new__(mcs, *args)

        if args[0] != 'ReplicaMixin':
            _MetaUtils.check_cqrs_id(model_cls)
            ReplicaMeta._check_cqrs_mapping(model_cls)

            ReplicaRegistry.register_model(model_cls)

        return model_cls

    @staticmethod
    def _check_cqrs_mapping(model_cls):
        """ Check that model has correct CQRS mapping configuration.

        :param dj_cqrs.mixins.ReplicaMixin model_cls: CQRS Replica Model.
        :raises: AssertionError
        """
        if model_cls.CQRS_MAPPING is not None:
            cqrs_field_names = list(model_cls.CQRS_MAPPING.values())
            _MetaUtils.check_cqrs_field_setting(model_cls, cqrs_field_names, 'CQRS_MAPPING')


class _MetaUtils:
    @classmethod
    def check_cqrs_field_setting(cls, model_cls, cqrs_field_names, cqrs_attr):
        cls._check_no_duplicate_names(model_cls, cqrs_field_names, cqrs_attr)
        cls._check_id_in_names(model_cls, cqrs_field_names, cqrs_attr)
        cls._check_unexisting_names(model_cls, cqrs_field_names, cqrs_attr)

    @staticmethod
    def check_cqrs_id(model_cls):
        """ Check that CQRS Model has CQRS_ID set up. """
        assert model_cls.CQRS_ID, 'CQRS_ID must be set for every model, that uses CQRS.'

    @staticmethod
    def _check_no_duplicate_names(model_cls, cqrs_field_names, cqrs_attr):
        model_name = model_cls.__name__

        assert len(set(cqrs_field_names)) == len(cqrs_field_names), \
            'Duplicate names in {} field for model {}.'.format(cqrs_attr, model_name)

    @staticmethod
    def _check_unexisting_names(model_cls, cqrs_field_names, cqrs_attr):
        opts = model_cls._meta
        model_name = model_cls.__name__

        model_field_names = {f.name for f in chain(opts.concrete_fields, opts.private_fields)}
        assert not set(cqrs_field_names) - model_field_names, \
            '{} field is not correctly set for model {}.'.format(cqrs_attr, model_name)

    @staticmethod
    def _check_id_in_names(model_cls, cqrs_field_names, cqrs_attr):
        opts = model_cls._meta
        model_name = model_cls.__name__

        pk_name = opts.pk.name
        assert pk_name in cqrs_field_names, \
            'PK is not in {} for model {}.'.format(cqrs_attr, model_name)



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

from django.db import models

from mptt.models import MPTTModel, TreeForeignKey
from dj_cqrs.constants import ALL_BASIC_FIELDS
from dj_cqrs.metas import MasterMeta
from dj_cqrs.mixins import MasterMixin, RawMasterMixin


class BasicFieldsModel(MasterMixin, models.Model):
    CQRS_ID = 'basic'

    int_field = models.IntegerField(primary_key=True)
    bool_field = models.NullBooleanField()
    char_field = models.CharField(max_length=200, null=True)
    date_field = models.DateField(null=True)
    datetime_field = models.DateTimeField(null=True)
    float_field = models.FloatField(null=True)
    url_field = models.URLField(null=True)
    uuid_field = models.UUIDField(null=True)


class AllFieldsModel(MasterMixin, models.Model):
    CQRS_FIELDS = ALL_BASIC_FIELDS
    CQRS_ID = 'all'

    int_field = models.IntegerField(null=True)
    char_field = models.CharField(max_length=200, null=True)


class ChosenFieldsModel(MasterMixin, models.Model):
    CQRS_FIELDS = ('char_field', 'id')
    CQRS_ID = 'chosen'

    float_field = models.IntegerField(null=True)
    char_field = models.CharField(max_length=200, null=True)


class AutoFieldsModel(MasterMixin, models.Model):
    CQRS_ID = 'auto'

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)


class SimplestModel(MasterMixin, models.Model):
    CQRS_ID = 'pk'

    id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=200, null=True)


class FilteredSimplestModel(MasterMixin, models.Model):
    CQRS_ID = 'filter'

    name = models.CharField(max_length=200)

    def is_sync_instance(self):
        return len(str(self.name)) > 2


class Publisher(models.Model):
    id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=20)


class Author(MasterMixin, models.Model):
    CQRS_ID = 'author'
    CQRS_SERIALIZER = 'tests.dj_master.serializers.AuthorSerializer'

    id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=20)

    publisher = models.ForeignKey(
        Publisher, related_name='authors', on_delete=models.SET_NULL, null=True,
    )

    @classmethod
    def relate_cqrs_serialization(cls, queryset):
        return queryset.select_related('publisher').prefetch_related('books')


class Book(models.Model):
    id = models.IntegerField(primary_key=True)
    title = models.CharField(max_length=20)

    author = models.ForeignKey(Author, related_name='books', on_delete=models.CASCADE)


class BadSerializationClassModel(MasterMixin, models.Model):
    CQRS_ID = 'bad_serialization'
    CQRS_SERIALIZER = 'tests.Serializer'


class BadQuerySetSerializationClassModel(MasterMixin, models.Model):
    CQRS_ID = 'bad_queryset'
    CQRS_SERIALIZER = 'tests.dj_master.serializers.AuthorSerializer'

    @classmethod
    def relate_cqrs_serialization(cls, queryset):
        return queryset.none()


class NonMetaClassModel(MPTTModel, RawMasterMixin):
    CQRS_ID = 'non_meta'

    name = models.CharField(max_length=50, unique=True)
    parent = TreeForeignKey(
        'self', on_delete=models.CASCADE, null=True, blank=True, related_name='children',
    )


MasterMeta.register(NonMetaClassModel)


class NonSentModel(MasterMixin, models.Model):
    CQRS_ID = 'non_sent'
    CQRS_PRODUCE = False



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

from django.conf import settings
from django.utils.module_loading import import_string

from dj_cqrs.transport.base import BaseTransport
from dj_cqrs.transport.rabbit_mq import RabbitMQTransport


transport_cls_location = getattr(settings, 'CQRS', {}) \
    .get('transport')
if not transport_cls_location:
    raise AttributeError('CQRS transport is not set.')

try:
    current_transport = import_string(transport_cls_location)

    if not issubclass(current_transport, BaseTransport):
        raise ValueError

except (ImportError, ValueError):
    raise ImportError('Bad CQRS transport class.')


__all__ = [BaseTransport, RabbitMQTransport, current_transport]



#  Copyright © 2020 Ingram Micro Inc. All rights reserved.

from dj_cqrs.controller.consumer import consume
from dj_cqrs.controller.producer import produce
from dj_cqrs.dataclasses import TransportPayload


def test_producer(mocker):
    transport_mock = mocker.patch('tests.dj.transport.TransportStub.produce')
    produce(TransportPayload('a', 'b', {}, 'c'))

    assert transport_mock.call_count == 1
    assert transport_mock.call_args[0][0].to_dict() == {
        'signal_type': 'a',
        'cqrs_id': 'b',
        'instance_data': {},
        'instance_pk': 'c',
    }


def test_consumer(mocker):
    factory_mock = mocker.patch('dj_cqrs.controller.consumer.route_signal_to_replica_model')
    consume(TransportPayload('a', 'b', {}, 'c'))

    factory_mock.assert_called_once_with('a', 'b', {})
